import os
import sys
import time
from argparse import ArgumentParser

import numpy as np
from soursop.sstrajectory import SSTrajectory

from starling.structure import ensemble
from starling.utilities import parse_output_path, symmetrize_distance_maps


## USER-FACING FUNCTIONS
def starling2xtc():
    # Initialize the argument parser
    parser = ArgumentParser(
        description="Convert .starling ensembles to a PDB topology file and XTC trajectory. Note if the structure reconstruction has not already been run, this will reconstruct a structural ensemble.")
    

    # Add command-line arguments corresponding to the parameters of the generate function
    parser.add_argument(
        "input_file", type=str, help="Input starling file", default=None
    )

    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=".",
        help="Directory and/or filename to save output (default: '.')",
    )

    parser.add_argument(
        "--device",
        type=str,
        default=None,
        help="Device we run conversion on (default: None)",
    )

    # Parse the command-line arguments
    args = parser.parse_args()

    outname = parse_output_path(args)

    # build an ensemble from the starling file
    E = ensemble.load_ensemble(args.input_file)

    # build with specified device
    E.build_ensemble_trajectory(device=args.device)

    # save
    E.save_trajectory(outname)


def starling2pdb():
    # Initialize the argument parser
    parser = ArgumentParser(description="Convert .starling files to a PDB trajectory. Note if the structure reconstruction has not already been run, this will reconstruct a structural ensemble.")

    # Add command-line arguments corresponding to the parameters of the generate function
    parser.add_argument(
        "input_file", type=str, help="Input starling file", default=None
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=".",
        help="Directory and/or filename to save output (default: '.')",
    )

    parser.add_argument(        
        "--device",
        type=str,
        default=None,
        help="Device we run conversion on (default: None)",
    )


    # Parse the command-line arguments
    args = parser.parse_args()
    outname = parse_output_path(args)

    # build an ensemble from the starling file
    E = ensemble.load_ensemble(args.input_file)

    # build with specified device
    E.build_ensemble_trajectory(device=args.device)

    # save the ensemble as a pdb trajectory
    E.save_trajectory(outname, pdb_trajectory=True)


def starling2numpy():
    # Initialize the argument parser
    parser = ArgumentParser(description="Convert .starling ensembles to a numpy array")

    # Add command-line arguments corresponding to the parameters of the generate function
    parser.add_argument(
        "input_file", type=str, help="Input starling file", default=None
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=".",
        help="Directory and/or filename to save output (default: '.')",
    )

    # Parse the command-line arguments
    args = parser.parse_args()

    outname = parse_output_path(args)

    # build an ensemble from the starling file
    E = ensemble.load_ensemble(args.input_file)

    dm = E.distance_maps()
    np.save(outname, dm)


def starling2sequence():
    # Initialize the argument parser
    parser = ArgumentParser(description="Convert .starling ensembles to a numpy array")

    # Add command-line arguments corresponding to the parameters of the generate function
    parser.add_argument(
        "input_file", type=str, help="Input starling file", default=None
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=".",
        help="Directory and/or filename to save output (default: '.')",
    )

    # Parse the command-line arguments
    args = parser.parse_args()

    # build an ensemble from the starling file
    E = ensemble.load_ensemble(args.input_file)

    print(E.sequence)


def starling2info():
    # Initialize the argument parser
    parser = ArgumentParser(description="Convert .starling ensembles to a numpy array")

    # Add command-line arguments corresponding to the parameters of the generate function
    parser.add_argument(
        "input_file", type=str, help="Input starling file", default=None
    )

    # Parse the command-line arguments
    args = parser.parse_args()

    # build an ensemble from the starling file
    E = ensemble.load_ensemble(args.input_file)

    
    if E.has_structures:
        marker = "[X] ({} structures)".format(len(E))        
    else:
        marker = "[ ]"
        

    print("-------------------------------")
    print("STARLING Generated ensemble")
    print("-------------------------------")
    print(f"Number of conformations     : {E.number_of_conformations}")
    print(f"Generate with STARLING      : {E._Ensemble__metadata['VERSION']}")
    print(f"Generate on ....            : {E._Ensemble__metadata['DATE']}")
    print(
        f"DDPM_WEIGHTS_PATH           : {E._Ensemble__metadata['DEFAULT_DDPM_WEIGHTS_PATH']}"
    )
    print(
        f"ENCODER_WEIGHTS_PATH        : {E._Ensemble__metadata['DEFAULT_ENCODER_WEIGHTS_PATH']}"
    )
    print(f"Average radius of gyration  : {E.radius_of_gyration(return_mean=True)}")
    print(f"Average end-to-end distance : {E.end_to_end_distance(return_mean=True)}")
    print(f"Sequence                    : {E.sequence}")
    print(f"Structures?                 : {marker}")
    print("-------------------------------")


def starling2starling():
    # Initialize the argument parser
    parser = ArgumentParser(
        description="Convert .starling ensembles to a different .starling ensemble! Right now this is for error correction."
    )

    # Add command-line arguments corresponding to the parameters of the generate function
    parser.add_argument(
        "input_file", type=str, help="Input starling file", default=None
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=".",
        help="Directory and/or filename to save output (default: '.')",
    )
    parser.add_argument(
        "--error-check",
        action="store_true",
        default=False,
        help="If set then the we do error checking to scan the ensemble for any issues.",
    )
    parser.add_argument(
        "--remove-errors",
        action="store_true",
        default=False,
        help="If set then the we output a new .starling file with any error-identified confomers removed.",
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        default=False,
        help="If set then the fixed file just overwrites the input file (including path information)...",
    )

    # Parse the command-line arguments
    args = parser.parse_args()

    outname = parse_output_path(args)

    # build an ensemble from the starling file
    E = ensemble.load_ensemble(args.input_file)

    if args.error_check:
        nframes = len(E)
        E.check_for_errors(remove_errors=args.remove_errors, verbose=True)
        if args.remove_errors and len(E) != nframes:
            if args.overwrite:
                outname = args.input_file
            else:
                outname = "fixed_" + outname
            E.save(outname, verbose=False)


def numpy2starling():
    # Initialize the argument parser
    parser = ArgumentParser(
        description="Convert nd.numpy array (generated by STARLING) files to a .starling file. Note this is ONLY to support backwards compatiblity for ensembles generated previously"
    )

    # Add command-line arguments corresponding to the parameters of the generate function
    parser.add_argument(
        "input_file", type=str, help="Input starling file", default=None
    )
    parser.add_argument(
        "-s",
        "--sequence",
        type=str,
        default=None,
        help="Amino acid sequence of the protein; this is a required argument",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=".",
        help="Directory and/or filename to save output (default: '.')",
    )
    parser.add_argument("-x", "--xtc", type=str, default=None, help="XTC trajectory")
    parser.add_argument("-p", "--pdb", type=str, default=None, help="PDB trajectory")
    parser.add_argument(
        "--build-structures",
        action="store_true",
        default=False,
        help="If set then the .starling file will include a reconstructred ensemble unless an XTC ensemble is passed in which case this is ignored.",
    )

    # Parse the command-line arguments
    args = parser.parse_args()

    if args.pdb is None and args.sequence is None:
        print("Error: EITHER sequence is a required or a PDB file is required")
        sys.exit(1)

    if args.xtc is not None and args.pdb is None:
        print("Error: PDB file is required when XTC file is provided")
        sys.exit(1)

    outname = parse_output_path(args)

    # build an ensemble from the starling file
    try:
        distance_maps = np.load(args.input_file)
    except Exception as e:
        print("Error loading numpy array: ", e)
        sys.exit(1)

    # if a pdb trajectory is provided, we can use it to get the sequence
    pdb_only = False
    try:
        if args.pdb and args.xtc is None:
            ssprot = SSTrajectory(args.pdb, args.pdb).proteinTrajectoryList[0]
            pdb_only = True
        elif args.pdb and args.xtc:
            ssprot = SSTrajectory(args.xtc, args.pdb).proteinTrajectoryList[0]
        else:
            ssprot = None
    except Exception as e:
        print("Error loading structural information: ", e)
        sys.exit(1)

    # if now sequence parsed get it from the parsed pdb trajectory
    if args.sequence is None:
        args.sequence = ssprot.get_amino_acid_sequence(oneletter=True)

    # overide this so we don't load a single conformer structure as the ensemble
    if pdb_only:
        ssprot = None

    # check sequence and distance map match in length
    seq_len = distance_maps.shape[1]
    if len(args.sequence) != seq_len:
        print("Error: sequence length does not match the distance map")
        sys.exit(1)

    # build
    E = ensemble.Ensemble(distance_maps, args.sequence, ssprot_ensemble=ssprot)

    if args.build_structures:
        E.build_ensemble_trajectory()

    # not ideal but we assign date to the file creation/touch time
    stat = os.stat(args.input_file)
    creation_time = stat.st_ctime
    DATE = time.ctime(creation_time)

    # finally we over-write the metadata
    E._Ensemble__metadata["DEFAULT_ENCODER_WEIGHTS_PATH"] = "UNKNOWN (from numpy array)"
    E._Ensemble__metadata["DEFAULT_DDPM_WEIGHTS_PATH"] = "UNKNOWN (from numpy array)"
    E._Ensemble__metadata["VERSION"] = "UNKNOWN (from numpy array)"
    E._Ensemble__metadata["DATE"] = DATE

    E.save(outname)


def xtc2starling():
    from pathlib import Path

    # Initialize the argument parser
    parser = ArgumentParser(description="Convert .xtc files to .starling ensembles")

    # Add command-line arguments corresponding to the parameters of the generate function
    parser.add_argument("--xtc", type=str, help="Input xtc file", default=None)
    parser.add_argument("--pdb", type=str, help="PDB topology file", default=None)
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=".",
        help="Directory and/or filename to save output (default: '.')",
    )

    # Parse the command-line arguments
    args = parser.parse_args()

    outname = Path(args.xtc).parts[-1].replace(".xtc", ".starling")
    ss_traj = SSTrajectory(args.xtc, args.pdb)
    ss_protein = ss_traj.proteinTrajectoryList[0]
    sequence = ss_protein.get_amino_acid_sequence(oneletter=True)
    distance_maps, _ = ss_protein.get_distance_map(return_instantaneous_maps=True)
    sym_distance_maps = symmetrize_distance_maps(distance_maps)

    # build an ensemble from the starling file
    ens = ensemble.Ensemble(sym_distance_maps, sequence)

    ens.save(outname)



from datetime import datetime

import numpy as np
import torch
from soursop.ssprotein import SSProtein
from soursop.sstrajectory import SSTrajectory
from tqdm import tqdm

from starling import configs, utilities
from starling._version import (
    __version__,
)
from starling.structure.coordinates import (
    create_ca_topology_from_coords,        
    generate_3d_coordinates_from_distances,
)


class Ensemble:
    """
    Class to represent an ensemble of conformations of a protein chain.
    The ensemble is represented by a 3D np.ndarray of N distance maps, where each
    distance map is a 2D numpy array.

    """

    def __init__(self, distance_maps, sequence, ssprot_ensemble=None):
        """
        Initialize the ensemble with a list of distance maps and the sequence
        of the protein chain.

        Parameters
        ----------
        distance_maps : np.ndarray
            3D Numpy array of shape (n_conformations, n_residues, n_residues).
            Note this this expects symmetrized distance maps.

        sequence : str
            Amino acid sequence of the protein chain.

        ssprot_ensemble : soursop.ssprotein.SSProtein
            SOURSOP SSProtein object. If provided, the ensemble will be initialized
            using this.

        """

        # sanity check input
        self.__sanity_check_init(distance_maps, sequence, ssprot_ensemble)

        self.__distance_maps = distance_maps
        self.sequence = sequence
        self.number_of_conformations = len(distance_maps)
        self.sequence_length = len(sequence)

        # initailize and then compute as needed
        self._rg_vals = []

        if ssprot_ensemble is None:
            self.__trajectory = None
        elif isinstance(ssprot_ensemble, SSProtein):
            self.__trajectory = ssprot_ensemble
        else:
            raise TypeError(
                "ssprot_ensemble must be a soursop.ssprotein.SSProtein object"
            )

        self.__metadata = {}
        self.__metadata["DEFAULT_ENCODER_WEIGHTS_PATH"] = (
            configs.DEFAULT_ENCODER_WEIGHTS_PATH
        )
        self.__metadata["DEFAULT_DDPM_WEIGHTS_PATH"] = configs.DEFAULT_DDPM_WEIGHTS_PATH
        self.__metadata["VERSION"] = __version__
        self.__metadata["DATE"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    def __sanity_check_init(self, distance_maps, sequence, ssprot_ensemble):
        """
        Perform sanity checks on the distance maps and sequence.

        Parameters
        ----------
        distance_maps : np.ndarray
            3D Numpy array of shape (n_conformations, n_residues, n_residues).

        sequence : str
            Amino acid sequence of the protein chain.

        Raises
        ------
        ValueError
            If the distance maps are not a list of 2D numpy arrays or if the
            sequence is not a string.

        """

        if not isinstance(distance_maps, np.ndarray):
            raise ValueError("distance_maps must be a numpy ndarray")

        if not all([isinstance(d, np.ndarray) and d.ndim == 2 for d in distance_maps]):
            raise ValueError("distance_maps must be a list of 2D numpy arrays")

        if not all(
            [
                d.shape[0] == d.shape[1] and d.shape[0] == len(sequence)
                for d in distance_maps
            ]
        ):
            raise ValueError(
                "distance_maps must be square matrices with the same size as the sequence"
            )

        if not isinstance(sequence, str):
            raise ValueError("sequence must be a string")

        if not all(char in configs.VALID_AA for char in sequence):
            raise ValueError(
                "sequence must contain only valid amino acid characters ({configs.VALID_AA})"
            )

        if ssprot_ensemble is not None:
            if not isinstance(ssprot_ensemble, SSProtein):
                raise ValueError(
                    "ssprot_ensemble must be a soursop.ssprotein.SSProtein object"
                )

    def check_for_errors(
        self, remove_errors=False, verbose=True, rebuild_trajectory=False
    ):
        """
        Function which scans the ensemble and finds any frames which may be erroneous
        based on impossible intermolecular distances.

        Note if the ensemble has an SSProtein object associated with it and remove_errors
        is set to true, this will either rebuild the trajectory object from scratch (if
        rebuild_trajectory=True) or delete the SSProtein object (if rebuild_trajectory=False).

        Parameters
        ------------
        remove_errors : bool
            If set to True, the erroneous frames are removed from the ensemble.

        verbose : bool
            If set to True, the function will print out the indices of the erroneous
            frames.

        rebuild_trajectory : bool
            If True, and if remove_errors set to True, AND if the ensemble has trajectory
            (SSProtein) object associated with it, this will trigger the reconstruction
            of this trajectory object with the removed frames. If set to False, and if
            remove_errors set to True, AND if the ensemble has a trajectory (SSProtein)
            object associated with it, this will delete that object.

        Parameters
        ------------
        list
            List of indices of the erroneous frames (note if they have been removed)
            these indices no longer make sense...

        """

        bad_frames = []
        for idx, distance_map in enumerate(self.__distance_maps):
            if utilities.check_distance_map_for_error(
                distance_map, min_separation=1, max_separation=4
            ):
                bad_frames.append(idx)

        if len(bad_frames) > 0:
            if verbose:
                print(f"Found {len(bad_frames)} bad frames: {bad_frames}")

            # remove frames and update any derived values
            if remove_errors:
                if verbose:
                    print("Removing bad frames")
                self.__distance_maps = np.delete(
                    self.__distance_maps, bad_frames, axis=0
                )
                self._rg_vals = []
                self.number_of_conformations = len(self.__distance_maps)

                if self.__trajectory is not None:
                    if rebuild_trajectory:
                        # delete and zero
                        self.build_ensemble_trajectory(force_recompute=True)

        return bad_frames

    def rij(self, i, j, return_mean=False):
        """
        Compute the distance between residues i and j for each conformation
        in the ensemble.

        Parameters
        ----------
        i : int
            Index of the first residue.

        j : int
            Index of the second residue.

        Returns
        -------
        list of float
            List of distances between residues i and j for each conformation
            in the ensemble. If return_mean is set returns the mean value.


        """
        if i < 0 or i >= self.sequence_length:
            raise ValueError(f"Invalid residue index i: {i}")

        tmp = []
        for d in self.__distance_maps:
            tmp.append(d[i][j])

        tmp = np.array(tmp)

        if return_mean:
            return np.mean(tmp)
        else:
            return tmp

    def end_to_end_distance(self, return_mean=False):
        """
        Compute the end-to-end distance of the protein chain
        for each conformation in the ensemble.

        Parameters
        ----------
        return_mean : bool
            If True, returns the mean end-to-end distance of the ensemble.

        Returns
        -------
        list of float
            List of end-to-end distances for each conformation in the ensemble.

        """

        tmp = self.rij(0, self.sequence_length - 1)

        if return_mean:
            return np.mean(tmp)
        else:
            return tmp

    def distance_maps(self, return_mean=False):
        """
        Return the collection of distance maps for the ensemble.

        Parameters
        ----------
        return_mean : bool
            If True, returns the mean distance map will be returned.
            Default is False.

        Returns
        -------
        np.array or list of np.array
            If return_mean is set to True, returns the mean distance map,
            otherwise returns the list of distance maps. Each distance map
            is a 2D numpy array.

        """
        if return_mean:
            return np.mean(self.__distance_maps, 0)
        else:
            return self.__distance_maps

    def contact_map(self, contact_thresh=11, return_mean=False, return_summed=False):
        """
        Return the collection of contact maps for the ensemble.
        Contacts are defined when residues are within a certain distance. Note
        only one of return_mean and return_summed can be set to True. If both
        are set to False, returns the array of instantaneous of contact maps.
        Else returns the averaged or the sum,ed of the contact maps.

        Parameters
        ----------
        contact_thresh : float
            Distance threshold for defining contacts. Default is 11
            Angstroms.

        return_mean : bool
            If True, the average contact map will be returned, meaning
            each element is between 0 and 1. Default is False.

        return_summed : bool
            If True, the summed contact map will be returned, meaning
            each element is an integer. Default is False.

        Returns
        -------
        np.array or list of np.array
            If return_mean is set to True, returns the mean distance map,
            otherwise returns the list of distance maps. Each distance map
            is a 2D numpy array.

        """
        # sanity check
        if return_mean and return_summed:
            raise ValueError("return_mean and return_summed cannot both be set to True")

        # get the distance maps
        dm = self.distance_maps(return_mean=False)

        if return_mean:
            return np.mean(np.array(dm < contact_thresh, dtype=int), 0)

        elif return_summed:
            return np.sum(np.array(dm < contact_thresh, dtype=int), 0)
        else:
            return np.array(dm < contact_thresh, dtype=int)

    def radius_of_gyration(
        self, return_mean=False, force_recompute=False, use_slow=False
    ):
        """
        Compute the radius of gyration of the protein chain
        for each conformation in the ensemble.

        Parameters
        ----------
        return_mean : bool
            If True, returns the mean radius of gyration of the ensemble.
            Default is False.

        force_recompute : bool
            If True, forces recomputation of the radius of gyration, otherwise
            uses the cached value if previously computed.
            Default is False.


        Returns
        -------
        np.array or float
            Array of radii of gyration for each conformation in the ensemble.
            If return_mean is set to true returns the mean value as a float

        """
        if len(self._rg_vals) == 0 or force_recompute == True:
            for d in self.__distance_maps:
                distances = np.sum(np.square(d))
                rg_val = np.sqrt(distances / (2 * np.power(self.sequence_length, 2)))
                self._rg_vals.append(rg_val)

            self._rg_vals = np.array(self._rg_vals)

        if return_mean:
            return np.mean(self._rg_vals)
        else:
            return self._rg_vals

    def local_radius_of_gyration(self, start, end, return_mean=False):
        """
        Return the local radius of gyration of the protein chain based on a subregion
        of the chain.

        Parameters
        ----------
        start : int
            The starting residue index.

        end : int
            The ending residue index.

        return_mean : bool
            If True, returns the mean radius of gyration of the ensemble.
            Default is False.

        Returns
        -------
        np.array or float
            Array of radii of gyration for each conformation in the ensemble.
            If return_mean is set to true returns the mean value as a float
        """

        local_rg = []
        for d in self.__distance_maps:
            distances = np.sum(np.square(d[start:end, start:end]))
            rg_val = np.sqrt(distances / (2 * np.power(end - start, 2)))
            local_rg.append(rg_val)

        local_rg = np.array(local_rg)

        if return_mean:
            return np.mean(local_rg)
        return local_rg

    def build_ensemble_trajectory(
        self,
        batch_size=100,
        num_cpus_mds=configs.DEFAULT_CPU_COUNT_MDS,
        num_mds_init=configs.DEFAULT_MDS_NUM_INIT,
        device=None,
        force_recompute=False,
        progress_bar=True,
    ):
        """
        Function that explicitly reconstructs a 3D ensemble of conformations
        using the distance maps. This happens automatically if the trajectory
        property is called, but this function allows for more control over the
        process. Specifically it allows you to specify the method used to generate
        the 3D structures, the number of CPUs to use, and the device to use for
        predictions. Note that if the 3D ensemble has already been reconstructed this
        function will NOT reconstructed the 3D ensemble unless force_recompute is set
        to True.

        Parameters
        ----------

        num_cpus_mds : int
            The number of CPUs to use for MDS. Default is 4 (set by
            configs.DEFAULT_CPU_COUNT_MDS)

        num_mds_init : int
            Number of independent MDS jobs to execute. NB: if this is
            increased this in principle means there are more chances
            of finding a good solution, but there is a performance hit
            unless num_cpus_mds >= num_mds_init. Default is
            4 (set by configs.DEFAULT_MDS_NUM_INIT).

        device : str
            The device to use for predictions. Default is None. If None, the
            default device will be used. Default device is 'gpu'.
            This is MPS for apple silicon and CUDA for all other devices.
            If MPS and CUDA are not available, automatically falls back to CPU.

        force_recompute : bool
            If True, forces recomputation of the ensemble trajectory, otherwise
            uses the cached trajectory if previously computed.
            Default is False.

        progress_bar : bool
            If True, displays a progress bar when generating the ensemble
            trajectory. Default is True.

        Returns
        -------
        soursop.sstrajectory.SSTrajectory
            The ensemble trajectory as a SOURSOP Trajectory object. Note that
            this object


        """

        # define and sanitize the device (we cast to string to ensure its a string cos 
        # generate_3d_coordinates_from_distances expects a string
        device = str(utilities.check_device(device))
        
        # if no traj yet or we're focing to recompute...
        if self.__trajectory is None or force_recompute:

            # build the 3D coordinates
            coordinates = generate_3d_coordinates_from_distances(device, batch_size, num_cpus_mds, num_mds_init, self.__distance_maps, progress_bar=progress_bar)

            # make an mdtraj.Trajectory object and then use that to initailize a SOURSOP SSTrajectory object
            self.__trajectory = SSTrajectory(
                TRJ=create_ca_topology_from_coords(self.sequence, coordinates)
            ).proteinTrajectoryList[0]

        return self.__trajectory

    @property
    def has_structures(self):
        """
        Check if the ensemble has 3D structures.

        Returns
        -------
        bool
            True if the ensemble has 3D structures, False otherwise.

        """
        if self.__trajectory is None:
            return False
        return True
    
    @property
    def trajectory(self):
        """
        Return the ensemble trajectory.

        Returns
        -------
        soursop.sstrajectory.SSTrajectory
            The ensemble trajectory as a SOURSOP Trajectory object.

        """
        if self.__trajectory is None:
            self.build_ensemble_trajectory()
        return self.__trajectory

    def save(
        self,
        filename_prefix,
        compress=False,
        reduce_precision=None,
        compression_algorithm="lzma",
        verbose=True,
    ):
        """
        Save the ensemble to a file in the STARLING format. Note this
        will add the .starling extension to the filename if not provided
        and will strip of any existing extension passed.

        Parameters
        ----------
        filename_prefix : str
            The name of the file to save the ensemble to, excluding
            file extensions which are added automatically. Note that if you
            provide a file extension it will be stripped off.

        compress : bool
            Whether to compress the file or not. Default is False.

        reduce_precision : bool
            Whether to reduce the precision of the distance map to a
            single decimal point and cast to float16 if possible.
            Default is None, and then sets to False if compression is
            False, but True if compression is True. However it can be
            manually over-ridden.

        compression_algorithm : str
            The compression algorithm to use. Options are 'gzip' and 'lzma'.
            `lzma` gives better compression if reduce_precision is set to True,
            but actually 'gzip' is better if reduce_precision is False. 'lzma'
            is also slower than 'gzip'. Default is 'lzma'.

        verbose : bool
            Flag to define how noisy we should be


        """
        utilities.write_starling_ensemble(
            self,
            filename_prefix,
            compress=compress,
            reduce_precision=reduce_precision,
            compression_algorithm=compression_algorithm,
            verbose=verbose,
        )

    def save_trajectory(self, filename_prefix, pdb_trajectory=False):
        """
        Save the ensemble trajectory to a file. This ONLY saves the
        3D structural ensemble but does not save the STARLING-generated
        distance maps. We recommend using save() instead to save the
        full STARLING object.

        Parameters
        ----------
        filename : str
            The name of the file to save the trajectory to, excluding
            file extensions which are added automatically.

        pdb_trajectory : bool
            If set to True, the output trajectory is ONLY saved as a PDB
            file. If set to false, it is saved as a single PDB structure
            for topology and then the actual trajectory as an XTC file.

        """

        traj = self.trajectory.traj

        if pdb_trajectory:
            traj.save_pdb(filename_prefix + ".pdb")
        else:
            traj[0].save_pdb(filename_prefix + ".pdb")
            traj.save_xtc(filename_prefix + ".xtc")

    def __len__(self):
        """
        Return the number of conformations in the ensemble.
        """
        return len(self.__distance_maps)

    def __str__(self):
        """
        Return a string representation of the ensemble.
        """
        if self.has_structures:        
            marker = "[X]"
        else:
            marker = "[ ]"
        return f"ENSEMBLE | len={len(self.sequence)}, ensemble_size={len(self)}, structures={marker}"

    def __repr__(self):
        """
        Return a string representation of the ensemble.
        """
        return self.__str__()


## ------------------------------------------ END OF CLASS DEFINITION


def load_ensemble(filename, ignore_structures=False):
    """
    Function to read in a STARLING ensemble from a file and return the
    STARLING ensemble object.

    Parameters
    ---------------
    filename : str
        The filename to read the ensemble from (should be a .starling
        file generated by STARLING)

    ignore_structures : bool
        If set to True, the function will discard structures that 
        are part of the .starling file. This can be useful if you
        don't need the structures as it slows loading to parse.
    """

    # note there's exception handling in the utilities.py file, and we automatically
    # detect the compression algorithm based on the file extension
    return_dict = utilities.read_starling_ensemble(filename)

    # make sure we can extract out the core components
    try:
        sequence = return_dict["sequence"]
        distance_maps = return_dict["distance_maps"]
        if ignore_structures:
            traj = None
        else:
            traj = return_dict["traj"]
        DEFAULT_ENCODER_WEIGHTS_PATH = return_dict["DEFAULT_ENCODER_WEIGHTS_PATH"]
        DEFAULT_DDPM_WEIGHTS_PATH = return_dict["DEFAULT_DDPM_WEIGHTS_PATH"]
        VERSION = return_dict["VERSION"]
        DATE = return_dict["DATE"]
    except Exception as e:
        raise Exception(
            f"Error parsing STARLING ensemble data: {filename} [error 2]; error {e}"
        )

    try:
        E = Ensemble(distance_maps, sequence, traj)
    except Exception as e:
        raise Exception(
            f"Error initializing STARLING ensemble: {filename} [error 3]; error {e}"
        )

    # finally we over-write the metadata
    E._Ensemble__metadata["DEFAULT_ENCODER_WEIGHTS_PATH"] = DEFAULT_ENCODER_WEIGHTS_PATH
    E._Ensemble__metadata["DEFAULT_DDPM_WEIGHTS_PATH"] = DEFAULT_DDPM_WEIGHTS_PATH
    E._Ensemble__metadata["VERSION"] = VERSION
    E._Ensemble__metadata["DATE"] = DATE

    return E



import os
from argparse import ArgumentParser
from collections import OrderedDict
from pathlib import Path

import h5py
import hdf5plugin
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import sparrow
import torch
import torch.nn.functional as F
from scipy.stats import pearsonr
from tabulate import tabulate
from tqdm import tqdm

from starling.models.diffusion import DiffusionModel
from starling.models.unet import UNetConditional
from starling.models.vae import VAE
from starling.samplers.ddim_sampler import DDIMSampler


def int_to_seq(int_seq):
    """
    Convert an integer sequence to a string sequence.
    """
    aa_to_int = {
        "0": 0,
        "A": 1,
        "C": 2,
        "D": 3,
        "E": 4,
        "F": 5,
        "G": 6,
        "H": 7,
        "I": 8,
        "K": 9,
        "L": 10,
        "M": 11,
        "N": 12,
        "P": 13,
        "Q": 14,
        "R": 15,
        "S": 16,
        "T": 17,
        "V": 18,
        "W": 19,
        "Y": 20,
    }
    reversed_dict = {v: k for k, v in aa_to_int.items()}
    seq = ""
    for i in int_seq:
        if i != 0:
            seq += str(reversed_dict[i])
    return seq


def symmetrize_distance_map(dist_map):
    # Ensure the distance map is 2D

    dist_map = dist_map.squeeze(0) if dist_map.dim() == 3 else dist_map

    # Create a copy of the distance map to modify
    sym_dist_map = dist_map.clone()

    # Replace the lower triangle with the upper triangle values
    mask_upper_triangle = torch.triu(torch.ones_like(dist_map), diagonal=1).bool()
    mask_lower_triangle = ~mask_upper_triangle

    # Set lower triangle values to be the same as the upper triangle
    sym_dist_map[mask_lower_triangle] = dist_map.T[mask_lower_triangle]

    # Set diagonal values to zero
    sym_dist_map.fill_diagonal_(0)

    return sym_dist_map.detach().cpu().numpy().squeeze()


def load_hdf5_compressed(file_path, keys_to_load=None):
    """
    Loads data from an HDF5 file.

    Parameters:
        - file_path (str): Path to the HDF5 file.
        - keys_to_load (list): List of keys to load. If None, loads all keys.
    Returns:
        - dict: Dictionary containing loaded data.
    """
    data_dict = {}
    with h5py.File(file_path, "r") as f:
        keys = keys_to_load if keys_to_load else f.keys()
        for key in keys:
            if key == "dm":
                data_dict[key] = f[key][...]
            else:
                data_dict[key] = f[key][...]
    return data_dict


def reconstruct(model, distance_maps):
    recon_dm, _ = model(distance_maps)
    recon_dm = recon_dm.detach().cpu().numpy().squeeze()

    return recon_dm


def get_errors(recon_dm, dm, mask):
    recon = F.mse_loss(recon_dm, dm, reduction="none")

    recon = recon * mask
    all_mse = recon.sum(axis=(1, 2)) / mask.sum(axis=(1, 2))

    all_mse = np.array([i.item() for i in all_mse])

    recon_bonds = [i.diagonal(offset=1) for i in recon]
    mask_bonds = [i.diagonal(offset=1) for i in mask]

    bonds_mse = np.array(
        [(i.sum() / j.sum()).item() for i, j in zip(recon_bonds, mask_bonds)]
    )

    return all_mse, bonds_mse


def read_input_file(file_path):
    """
    Read the input file and return a list of paths to the HDF5 files.
    """
    paths = {}
    with open(file_path, "r") as f:
        for line in f:
            line = line.strip().split("\t")
            paths[line[0]] = line[1]
    return paths


def prepare_data(data):
    """
    Prepare the data for inference.
    """
    dm = torch.from_numpy(data)
    dm = dm.unsqueeze(1)

    return dm


def hellinger_distance(p, q):
    return np.sqrt(np.sum((np.sqrt(p) - np.sqrt(q)) ** 2)) / np.sqrt(2)


def hellingers_summary(
    gt_dms, recon_dms, save_path="hellinger_heatmap.png", plot_histogram=False
):
    """
    Compute Hellinger distances between corresponding residue pairs in ground truth and reconstructed distance maps.

    Parameters
    ----------
    gt_dms : np.ndarray
        Ground truth distance maps of shape (num_gt_conformers, padded_length, padded_length)
        where padded_length is the padded number of residues in the sequence
    recon_dms : np.ndarray
        Reconstructed distance maps of shape (num_recon_conformers, actual_length, actual_length)
        where actual_length is the actual number of residues in the sequence

    Returns
    -------
    np.ndarray
        Array of shape (actual_length, actual_length) containing Hellinger distances for each residue pair
    """
    actual_length = recon_dms.shape[1]

    # Trim ground truth maps to match the size of reconstructed maps
    gt_dms_trimmed = gt_dms[:, :actual_length, :actual_length]

    assert gt_dms_trimmed.shape[1:] == recon_dms.shape[1:], (
        "Inconsistent shapes between trimmed ground truth and reconstructed distance maps"
    )

    colors = ["#4c72b0", "#dd8452"]

    n_bins = 25
    # Initialize the result array
    hellinger_distances = np.zeros((actual_length, actual_length))
    gt_histograms = []
    recon_histograms = []
    for i in range(actual_length):
        for j in range(actual_length):
            global_min = min(gt_dms_trimmed[:, i, j].min(), recon_dms[:, i, j].min())

            global_max = max(gt_dms_trimmed[:, i, j].max(), recon_dms[:, i, j].max())

            bin_width = (global_max - global_min) / n_bins

            gt_values = gt_dms_trimmed[:, i, j]
            recon_values = recon_dms[:, i, j]

            hist_gt, _ = np.histogram(
                gt_values,
                bins=n_bins,
                range=(global_min, global_max),
                density=True,
            )

            hist_gt *= bin_width
            gt_histograms.append(hist_gt)
            hist_recon, _ = np.histogram(
                recon_values,
                bins=n_bins,
                range=(global_min, global_max),
                density=True,
            )
            hist_recon *= bin_width
            recon_histograms.append(hist_recon)
            os.makedirs("plots", exist_ok=True)
            if plot_histogram:
                # Plot histogram for ground truth values (gt_values)
                plt.hist(
                    gt_values,
                    bins=n_bins,
                    range=(global_min, global_max),
                    density=True,
                    alpha=0.6,  # Slightly increase the opacity
                    color=colors[0],  # Use custom color
                    edgecolor="black",  # Add edge color for visibility
                    label="Ground Truth",  # Label for legend
                    linewidth=1.2,  # Line width for the edges
                )

                # Plot histogram for reconstructed values (recon_values)
                plt.hist(
                    recon_values,
                    bins=n_bins,
                    range=(global_min, global_max),
                    density=True,
                    alpha=0.6,
                    color=colors[1],
                    edgecolor="black",
                    label="Reconstruction",
                    linewidth=1.2,
                )

                # Add labels, title, and legend
                plt.xlabel("Value", fontsize=14)
                plt.ylabel("Density", fontsize=14)
                plt.title(f"Residue Histogram for {i}-{j}", fontsize=16)
                plt.legend(fontsize=12)

                # Adjust ticks and add gridlines for better readability
                plt.xticks(fontsize=12)
                plt.yticks(fontsize=12)
                plt.grid(True, which="both", linestyle="--", linewidth=0.5, alpha=0.7)

                # Save the plot with high DPI and tight layout
                plt.savefig(
                    f"plots/residue_{i}_{j}_histogram.png", dpi=300, bbox_inches="tight"
                )

                # Close the figure to free memory
                plt.close()

            # Compute Hellinger distance
            hellinger_distances[i, j] = hellinger_distance(hist_gt, hist_recon)

    # Create and save the heatmap
    plt.figure(figsize=(10, 8))
    sns.heatmap(hellinger_distances, cmap="viridis", annot=False, cbar=True)
    plt.title(
        "Hellinger Distances between Ground Truth and Reconstructed Distance Maps"
    )
    plt.xlabel("Residue Index")
    plt.ylabel("Residue Index")
    plt.savefig(save_path, dpi=300, bbox_inches="tight")
    plt.close()

    return hellinger_distances


def main():
    parser = ArgumentParser()
    parser.add_argument("--input", type=str, required=True)
    parser.add_argument("--vae", type=str, required=True)
    parser.add_argument("--ddpm", type=str, required=True)
    parser.add_argument("--ddim", action="store_true")
    parser.add_argument("--steps", type=int, default=50)
    parser.add_argument("--conformations", type=int, default=100)
    parser.add_argument("--batch", type=int, default=100)
    parser.add_argument("--device", type=str, default="cuda:0")
    parser.add_argument("--outfile", type=str, default="summary_stats_ddpm.csv")
    args = parser.parse_args()

    # Load the VAE model
    vae = VAE.load_from_checkpoint(args.vae, map_location=args.device)

    # Load the UNet model
    UNet_model = UNetConditional(
        in_channels=1,
        out_channels=1,
        base=64,
        norm="group",
        blocks=[2, 2, 2],
        middle_blocks=2,
        labels_dim=384,
    )

    # Load the DDPM model
    diffusion = DiffusionModel.load_from_checkpoint(
        args.ddpm,
        model=UNet_model,
        encoder_model=vae,
        map_location=args.device,
    )

    # Construct a sampler
    if args.ddim:
        sampler = DDIMSampler(ddpm_model=diffusion, n_steps=args.steps)
    else:
        sampler = diffusion

    # Read the input file
    paths = read_input_file(args.input)

    # Start a dataframe
    results = OrderedDict()

    for path in tqdm(paths):
        sequence_stats = OrderedDict()

        if Path(paths[path]).suffix == ".h5":
            data = load_hdf5_compressed(paths[path], keys_to_load=["dm", "seq"])
            mpipi_dm = data["dm"]
            sequence = int_to_seq(data["seq"])
        else:
            mpipi_dm = []
            sequence = paths[path]

        num_batches = args.conformations // args.batch
        remaining_samples = args.conformations % args.batch

        starling_dm = []

        for batch in range(num_batches):
            distance_map = sampler.sample(args.batch, labels=sequence)
            starling_dm.append(
                [
                    symmetrize_distance_map(dm[:, : len(sequence), : len(sequence)])
                    for dm in distance_map
                ]
            )

        if remaining_samples > 0:
            distance_map = sampler.sample(remaining_samples, labels=sequence)
            starling_dm.append(
                [
                    symmetrize_distance_map(dm[:, : len(sequence), : len(sequence)])
                    for dm in distance_map
                ]
            )
        starling_dm = np.concatenate(starling_dm, axis=0)

        hellinger_distances = hellingers_summary(
            mpipi_dm, starling_dm, save_path=f"plots/hellinger_heatmap_{path}.png"
        )

        upper_triangle_bond_offset = np.triu(hellinger_distances, k=1).mean()

        albatross_prediction = sparrow.Protein(sequence).predictor.end_to_end_distance()
        starling_prediction = starling_dm.mean(axis=0)[0, -1]

        if len(mpipi_dm) > 0:
            mpipi_prediction = round(mpipi_dm.mean(axis=0)[0, len(sequence) - 1], 4)
            mean_mpipi = mpipi_dm[:, : len(sequence), : len(sequence)].mean(axis=0)
            mean_starling = starling_dm.mean(axis=0)
            dm_difference = abs(mean_starling - mean_mpipi)

            dm_difference_mean = round(dm_difference.mean(), 4)
            dm_difference_max = round(dm_difference.max(), 4)
        else:
            mpipi_prediction = None
            dm_difference_mean = None
            dm_difference_max = None

        sequence_stats["STARLING"] = round(starling_prediction, 4)
        sequence_stats["ALBATROSS"] = round(albatross_prediction, 4)
        sequence_stats["Mpipi"] = mpipi_prediction
        sequence_stats["Average_dm_abe"] = dm_difference_mean
        sequence_stats["Max_dm_abe"] = dm_difference_max
        sequence_stats["Mean_hellinger_distance"] = round(upper_triangle_bond_offset, 4)

        results[path] = sequence_stats

    results_df = pd.DataFrame(results).T

    correlation = {
        "STARLING": round(
            pearsonr(
                results_df["STARLING"],
                results_df["ALBATROSS"]
                if mpipi_prediction is None
                else results_df["Mpipi"],
            )[0],
            4,
        )
    }

    if len(mpipi_dm) > 0:
        correlation["ALBATROSS"] = round(
            pearsonr(
                results_df["ALBATROSS"],
                results_df["Mpipi"],
            )[0],
            4,
        )

    results_df.loc["Correlation"] = correlation

    results_df.to_csv(args.outfile, index=True)

    summary = results_df.tail(1).reset_index(drop=False)

    formatted_summary = tabulate(
        summary, headers="keys", tablefmt="pipe", floatfmt=".4f", showindex=False
    )

    print(formatted_summary)


if __name__ == "__main__":
    main()



from typing import Iterator, Tuple

import h5py
import hdf5plugin
import numpy as np
import pandas as pd


def one_hot_encode(sequences):
    """
    One-hot encodes a sequence.
    """
    # Define the mapping of each amino acid to a unique integer
    aa_to_int = {
        "0": 0,
        "A": 1,
        "C": 2,
        "D": 3,
        "E": 4,
        "F": 5,
        "G": 6,
        "H": 7,
        "I": 8,
        "K": 9,
        "L": 10,
        "M": 11,
        "N": 12,
        "P": 13,
        "Q": 14,
        "R": 15,
        "S": 16,
        "T": 17,
        "V": 18,
        "W": 19,
        "Y": 20,
    }

    # Convert sequences to numpy array of integers
    int_sequences = [[aa_to_int[aa] for aa in sequence] for sequence in sequences]
    int_sequences = np.array(int_sequences)

    # Create an array of zeros with shape (num_sequences, sequence_length, num_classes)
    num_sequences = len(sequences)
    sequence_length = max(len(sequence) for sequence in sequences)
    num_classes = len(aa_to_int)

    one_hot_encoded_seq = np.zeros(
        (num_sequences, sequence_length, num_classes), dtype=np.float32
    )

    # Use advanced indexing to set the appropriate elements to 1
    for i, sequence in enumerate(int_sequences):
        one_hot_encoded_seq[i, np.arange(len(sequence)), sequence] = 1

    return one_hot_encoded_seq


def MaxPad(original_array: np.array, shape: tuple) -> np.array:
    """
    A function that takes in a distance map and pads it to a desired shape

    Parameters
    ----------
    original_array : np.array
        A distance map

    Returns
    -------
    np.array
        A distance map padded to a desired shape
    """
    # Pad the distance map to a desired shape
    pad_height = max(0, shape[0] - original_array.shape[0])
    pad_width = max(0, shape[1] - original_array.shape[1])
    return np.pad(
        original_array,
        ((0, pad_height), (0, pad_width)),
        mode="constant",
        constant_values=0,
    )


def load_hdf5_compressed(file_path, frame=None, keys_to_load=None):
    """
    Loads data from an HDF5 file, optionally for a specific frame.

    Parameters:
        - file_path (str): Path to the HDF5 file.
        - frame (int, optional): Specific frame to load from datasets. If None, loads the entire dataset.
        - keys_to_load (list, optional): List of keys to load. If None, loads all keys.

    Returns:
        - dict: Dictionary containing loaded data with keys as dataset names and values as dataset contents.
    """
    data = {}
    with h5py.File(file_path, "r") as f:
        keys = keys_to_load if keys_to_load else list(f.keys())
        for key in keys:
            if frame is not None and key == "dm":
                data[key] = f[key][frame]
            else:
                data[key] = f[key][...]
    return data


def read_tsv_file(tsv_file: str) -> Iterator[Tuple[str, str]]:
    """
    A function that reads the paths to distance maps from a tsv file

    Parameters
    ----------
    tsv_file : str
        A path to a tsv file containing the paths to distance maps as a first column
        and index of a distance map to load as a second column

    Returns
    -------
    Iterator[Tuple[str, str]]
        An iterator of tuples containing paths to distance maps and their indices
    """
    df = pd.read_csv(tsv_file, sep="\t", header=None, usecols=[0, 1])
    return df


def symmetrize(matrix):
    """
    Symmetrizes a matrix.
    """
    if np.array_equal(matrix, matrix.T):
        return matrix
    else:
        # Extract upper triangle excluding diagonal
        upper_triangle = np.triu(matrix, k=1)
        # Symmetrize upper triangle by mirroring
        sym_matrix = upper_triangle + upper_triangle.T
        # Add diagonal elements (to handle odd-sized matrices)
        sym_matrix += np.diag(np.diag(matrix))
        return sym_matrix



import torch
import torch.nn as nn


class EMA:
    def __init__(self, beta: float = 0.999):
        self.beta = beta
        self.step = 0

    def update_model_average(self, ma_model, current_model):
        for current_params, ma_params in zip(
            current_model.parameters(), ma_model.parameters()
        ):
            ema_model_weights, current_weighs = ma_params.data, current_params.data
            ma_params.data = self.update_average(ema_model_weights, current_weighs)

    def update_average(self, ema_model_params, current_model_params):
        if ema_model_params is None:
            return current_model_params

        return ema_model_params * self.beta + (1 - self.beta) * current_model_params

    def step_ema(self, ema_model, model, step_start_ema=2000):
        if self.step < step_start_ema:
            self.reset_parameters(ema_model, model)
            self.step += 1
            return
        self.update_model_average(ema_model, model)
        self.step += 1

    def reset_parameters(self, ema_model, model):
        ema_model.load_state_dict(model.state_dict())



import math
from typing import List, Union

import pytorch_lightning as pl
import torch
import torch.nn as nn
from torch.amp import autocast
from torch.functional import F
from torch.optim.lr_scheduler import (
    CosineAnnealingLR,
    CosineAnnealingWarmRestarts,
    LambdaLR,
    OneCycleLR,
)
from tqdm import tqdm

from starling.data.data_wrangler import one_hot_encode
from starling.data.schedulers import (
    cosine_beta_schedule,
    linear_beta_schedule,
    sigmoid_beta_schedule,
)

# Adapted from https://github.com/Camaltra/this-is-not-real-aerial-imagery/blob/main/src/ai/diffusion_process.py
# and https://github.com/lucidrains/denoising-diffusion-pytorch/blob/main/denoising_diffusion_pytorch/classifier_free_guidance.py#L720

torch.set_float32_matmul_precision("high")


# Helper function
def extract(
    constants: torch.Tensor, timestamps: torch.Tensor, shape: int
) -> torch.Tensor:
    """
    A function to extract values from a tensor based on the timestamps

    Parameters
    ----------
    constants : torch.Tensor
        A tensor to extract values from
    timestamps : torch.Tensor
        A 1D tensor containing the indices to extract from the constants tensor
    shape : int
        Desired shape of the output tensor

    Returns
    -------
    torch.Tensor
        Returns the extracted values from the constants tensor
    """
    batch_size = timestamps.shape[0]
    out = constants.gather(-1, timestamps)
    return out.reshape(batch_size, *((1,) * (len(shape) - 1))).to(timestamps.device)


class DiffusionModel(pl.LightningModule):
    SCHEDULER_MAPPING = {
        "linear": linear_beta_schedule,
        "cosine": cosine_beta_schedule,
        "sigmoid": sigmoid_beta_schedule,
    }

    def __init__(
        self,
        model: nn.Module,
        encoder_model: nn.Module,
        image_size: int,
        *,
        beta_scheduler: str = "cosine",
        timesteps: int = 1000,
        schedule_fn_kwargs: Union[dict, None] = None,
        labels: str = "learned-embeddings",
        set_lr: float = 1e-4,
        config_scheduler: str = "LinearWarmupCosineAnnealingLR",
    ) -> None:
        """
        Denoising-diffusion model framework for latent space diffusion models. This
        class is a PyTorch Lightning module that can be used to train a diffusion model
        on a given dataset. The model can be used to sample from the trained model
        and generate new samples.

        References
        ----------
        1) Sohl-Dickstein, J., Weiss, E., Maheswaranathan, N. & Ganguli, S.
        Deep Unsupervised Learning using Nonequilibrium Thermodynamics.
        in Proceedings of the 32nd International Conference on Machine Learning
        (eds. Bach, F. & Blei, D.) vol. 37 2256–2265 (PMLR, Lille, France, 07--09 Jul 2015).

        2) Ho, J., Jain, A. & Abbeel, P. Denoising Diffusion Probabilistic Models. arXiv [cs.LG] (2020).

        3) Rombach, R., Blattmann, A., Lorenz, D., Esser, P. & Ommer, B.
        High-resolution image synthesis with latent diffusion models. arXiv [cs.CV] (2021).


        Parameters
        ----------
        model : nn.Module
            A U-Net model that takes in an image, a timestamp, and optionally labels to condition on
            and outputs the predicted noise
        encoder_model : nn.Module
            A VAE model that takes in an image (distance map) and outputs the latent space that the
            diffusion model is trained in
        image_size : int
            The size of the latent space (height and width)
        beta_scheduler : str, optional
            The name of the beta scheduler to use, by default "cosine"
        timesteps : int, optional
            The number of timesteps to run the diffusion process, by default 1000
        schedule_fn_kwargs : Union[dict, None], optional
            Additional arguments to pass to the beta scheduler function, by default None
        labels : str, optional
            The type of labels to condition the model on, by default "learned-embeddings"
        set_lr : float, optional
            The initial learning rate for the optimizer, by default 1e-4
        config_scheduler : str, optional
            The name of the learning rate scheduler to use, by default "CosineAnnealingLR"

        Raises
        ------
        ValueError
            If the beta scheduler is not implemented
        """
        super().__init__()

        # Save the hyperparameters of the model but ignore the encoder_model and the U-Net model
        self.save_hyperparameters(ignore=["encoder_model", "model"])

        self.model = model
        self.labels = labels

        # Freeze the encoder model parameters we don't want to keep training it (should already be trained)
        self.encoder_model = encoder_model
        for param in self.encoder_model.parameters():
            param.requires_grad = False
        self.encoder_model.eval()

        self.in_channels = self.model.in_channels
        self.image_size = image_size

        # Learning rate params
        self.set_lr = set_lr
        self.config_scheduler = config_scheduler

        self.beta_scheduler_fn = self.SCHEDULER_MAPPING.get(beta_scheduler)
        if self.beta_scheduler_fn is None:
            raise ValueError(f"unknown beta schedule {beta_scheduler}")

        if schedule_fn_kwargs is None:
            schedule_fn_kwargs = {}

        # This will be calculated later on during the first global step
        # Need to register here so pytorch_lightning doesn't freak out
        # Assuming the expected shape for the buffer is [1]
        # This is used to scale the latent space to have unit variance (see reference #3)
        latent_space_scaling_factor = torch.tensor(1.0, dtype=torch.float32)

        # Register the buffer
        self.register_buffer("latent_space_scaling_factor", latent_space_scaling_factor)

        betas = self.beta_scheduler_fn(timesteps, **schedule_fn_kwargs)
        alphas = 1.0 - betas
        alphas_cumprod = torch.cumprod(alphas, dim=0)
        alphas_cumprod_prev = F.pad(alphas_cumprod[:-1], (1, 0), value=1.0)
        posterior_variance = (
            betas * (1.0 - alphas_cumprod_prev) / (1.0 - alphas_cumprod)
        )

        # Register the buffers for the model
        self.register_buffer("betas", betas)
        self.register_buffer("alphas_cumprod", alphas_cumprod)
        self.register_buffer("alphas_cumprod_prev", alphas_cumprod_prev)
        self.register_buffer("sqrt_recip_alphas", torch.sqrt(1.0 / alphas))
        self.register_buffer("sqrt_alphas_cumprod", torch.sqrt(alphas_cumprod))
        self.register_buffer(
            "sqrt_one_minus_alphas_cumprod", torch.sqrt(1.0 - alphas_cumprod)
        )
        self.register_buffer("posterior_variance", posterior_variance)

        timesteps, *_ = betas.shape
        self.num_timesteps = int(timesteps)

        self.sampling_timesteps = timesteps

        self.monitor = "epoch_val_loss"

        if self.labels == "learned-embeddings":
            self.sequence_embedding = nn.Embedding(21, self.model.labels_dim)

    @torch.inference_mode()
    def p_sample(
        self, x: torch.Tensor, timestamp: int, labels: torch.Tensor
    ) -> torch.Tensor:
        """
        One denoising step of the diffusion model. This function is
        used in p_sample_loop to denoise the initial tensor sampled from N(0, I)

        Parameters
        ----------
        x : torch.Tensor
            A tensor to denoise
        timestamp : int
            The timestep of the denoising-diffusion process to denoise
        labels : torch.Tensor
            Labels (sequences) to condition the model on

        Returns
        -------
        torch.Tensor
            Returns the denoised tensor (t-1)
        """
        b, *_, device = *x.shape, x.device

        # Batch the timestep to the same size as the input tensor x
        batched_timestamps = torch.full(
            (b,), timestamp, device=device, dtype=torch.long
        )

        # Run the model to predict the noise at the current timestamp
        preds = self.model(x, batched_timestamps, labels)

        # Extract the necessary values from the buffers to calculate the predicted mean
        betas_t = extract(self.betas, batched_timestamps, x.shape)
        sqrt_recip_alphas_t = extract(
            self.sqrt_recip_alphas, batched_timestamps, x.shape
        )
        sqrt_one_minus_alphas_cumprod_t = extract(
            self.sqrt_one_minus_alphas_cumprod, batched_timestamps, x.shape
        )

        # Calculate the predicted mean based on the model prediction of the noise
        predicted_mean = sqrt_recip_alphas_t * (
            x - betas_t * preds / sqrt_one_minus_alphas_cumprod_t
        )

        # If the timestamp is 0, return the predicted mean
        if timestamp == 0:
            return predicted_mean
        else:
            posterior_variance = extract(
                self.posterior_variance, batched_timestamps, x.shape
            )
            noise = torch.randn_like(x)
            return predicted_mean + torch.sqrt(posterior_variance) * noise

    @torch.inference_mode()
    def p_sample_loop(
        self,
        shape: tuple,
        labels: torch.Tensor,
        steps: int = None,
        return_all_timesteps: bool = False,
    ) -> torch.Tensor:
        """
        Sampling loop for the diffusion model. It loops over the timesteps
        to denoise the initial tensor sampled from N(0, I)

        Parameters
        ----------
        shape : tuple
            The shape of the tensor to sample from N(0, I)
        labels : torch.Tensor
            Labels to condition the sampling on
        steps : int, optional
            Number of steps to sample, by default None (all timesteps)
        return_all_timesteps : bool, optional
            Whether to return the full trajectory of denoising,
            by default False

        Returns
        -------
        torch.Tensor
            Returns the fully denoised tensor, in this case a latent space
            that can be decoded using a pre-trained VAE
        """

        batch, device = shape[0], self.device

        all_latents = []

        # Sample noise to generate the latent representation of the data
        latents = torch.randn(shape, device=device)

        # Get the number of steps to run denoising
        if steps is not None:
            timesteps = (
                torch.linspace(1, self.num_timesteps - 1, steps).round().to(torch.int64)
            )
        else:
            timesteps = range(0, self.num_timesteps)

        # Loop over the timesteps to denoise the initial tensor
        for t in tqdm(
            reversed(timesteps),
            desc="Generating Latents",
            total=len(timesteps),
        ):
            if return_all_timesteps:
                all_latents.append(latents)

            latents = self.p_sample(latents, t, labels)

        # Scale the latents back to the original scale
        latents = (1 / self.latent_space_scaling_factor) * latents

        # if not return_all_timesteps else torch.stack(imgs, dim=1)
        if return_all_timesteps:
            return torch.stack(all_latents, dim=1) * (
                1 / self.latent_space_scaling_factor
            )
        else:
            return latents

    @torch.inference_mode()
    def sample(
        self,
        batch_size: int,
        labels: str,
        steps: int = None,
        return_all_timesteps: bool = False,
    ) -> torch.Tensor:
        """
        Sample from the trained diffusion model

        Parameters
        ----------
        batch_size : int
            The batch size to sample, higher the better if enough VRAM
        labels : str
            Sequence to condition the sampling on
        steps : int, optional
            Number of steps to sample, by default None (all timesteps)
        return_all_timesteps : bool, optional
            Whether to return all the tensors along the
            denoising trajectory, by default False

        Returns
        -------
        torch.Tensor
            Returns the sampled fully denoised tensor
        """

        shape = (batch_size, self.in_channels, self.image_size, self.image_size)

        # Prepare the labels to condition the sampling from the model on
        with torch.no_grad():
            if self.labels == "learned-embeddings":
                # Get the learned embeddings for the given sequence
                labels = (
                    torch.argmax(
                        torch.from_numpy(one_hot_encode(labels.ljust(384, "0"))), dim=-1
                    )
                    .to(torch.int64)
                    .squeeze()
                    .to(self.device)
                )
                labels = self.sequence2labels(labels)

        labels = labels.unsqueeze(0)

        # Sample the denoising-diffusion model to generate data (in our case, latent space)
        latents = self.p_sample_loop(
            shape, labels, steps=steps, return_all_timesteps=return_all_timesteps
        )

        # Decode the latent space with the VAE to generate the distance map
        distance_map = self.encoder_model.decode(latents)

        return distance_map, latents, labels

    # Remove mixed precision from this function, I've experienced numerical instability here
    @autocast(device_type="cuda", enabled=False)
    def q_sample(
        self, x_start: torch.Tensor, t: int, noise: torch.Tensor = None
    ) -> torch.Tensor:
        """
        Add the noise to x_start tensor based on the timestamp t

        Parameters
        ----------
        x_start : torch.Tensor
            The starting image tensor
        t : int
            The timestep of the denoising-diffusion process
        noise : torch.Tensor, optional
            Sampled noise to add, by default None

        Returns
        -------
        torch.Tensor
            Returns the properly (according to the timestamp) noised tensor
        """
        if noise is None:
            noise = torch.randn_like(x_start)

        # Extract the necessary values from the buffers to calculate the noise to be added
        sqrt_alphas_cumprod_t = extract(self.sqrt_alphas_cumprod, t, x_start.shape)
        sqrt_one_minus_alphas_cumprod_t = extract(
            self.sqrt_one_minus_alphas_cumprod, t, x_start.shape
        )

        # Return the noised tensor based on the timestamp
        return sqrt_alphas_cumprod_t * x_start + sqrt_one_minus_alphas_cumprod_t * noise

    def sequence2labels(self, sequences: List) -> torch.Tensor:
        """
        Converts sequences to labels based on user defined models,

        Parameters
        ----------
        sequences : List
            A list of sequences to convert to labels

        Returns
        -------
        torch.Tensor
            Returns the labels for the decoder

        Raises
        ------
        ValueError
            If the labels are not one of the three options
        """
        if self.labels == "learned-embeddings":
            encoded = self.sequence_embedding(sequences)

        return encoded

    def p_loss(
        self,
        x_start: torch.Tensor,
        t: int,
        labels: torch.Tensor = None,
        noise: torch.Tensor = None,
        loss_type: str = "l2",
    ) -> torch.Tensor:
        """
        A function that runs the model and calculates the loss based on the
        predicted noise and the actual noise. The loss can either be L1 or L2.

        Parameters
        ----------
        x_start : torch.Tensor
            The starting image tensor
        t : int
            The timestep along the denoising-diffusion process
        labels : torch.Tensor, optional
            Labels to condition the model on, by default None
        noise : torch.Tensor, optional
            Sampled noise from N(0,I), by default None
        loss_type : str, optional
            The type of loss to calculate between the
            amount of added noise and predicted noise, by default "l2"

        Returns
        -------
        torch.Tensor
            Returns the loss

        Raises
        ------
        ValueError
            If the loss type is not one of the two options (l1, l2)
        """
        if noise is None:
            noise = torch.randn_like(x_start)
            # Offset noise that seems to improve the inference
            # According to https://www.crosslabs.org/blog/diffusion-with-offset-noise
            # noise += 0.1 * torch.randn(
            #     x_start.shape[0], x_start.shape[1], 1, 1, device=self.device
            # )

        # Noise the input data
        x_noised = self.q_sample(x_start, t, noise=noise)

        # Get the labels to condition the model on
        labels = self.sequence2labels(labels)

        # Run the model to predict the noise
        predicted_noise = self.model(x_noised, t, labels)

        # Calculate the loss based on the predicted noise and the actual noise
        if loss_type == "l2":
            loss = F.mse_loss(noise, predicted_noise)
        elif loss_type == "l1":
            loss = F.l1_loss(noise, predicted_noise)
        else:
            raise ValueError(f"unknown loss type {loss_type}")

        return loss

    def forward(self, x: torch.Tensor, labels: torch.Tensor = None) -> torch.Tensor:
        """
        Forward pass of the model, calculates the loss based on the
        predicted noise and the actual noise.

        Parameters
        ----------
        x : torch.Tensor
            The starting tensor to noise/denoise
        labels : torch.Tensor, optional
            Sequences to condition the model on, by default None

        Returns
        -------
        torch.Tensor
            Returns the loss
        """
        b, c, h, w, device, img_size = *x.shape, x.device, self.image_size
        assert h == w == img_size, f"image size must be {img_size}"

        # Generate random timestamps to noise the tensor and learn the denoising process
        timestamps = torch.randint(0, self.num_timesteps, (b,), device=device).long()

        return self.p_loss(x, timestamps, labels)

    def training_step(self, batch: torch.Tensor, batch_idx: int) -> torch.Tensor:
        data, sequences = batch

        with torch.no_grad():
            latent_encoding = self.encoder_model.encode(data)
            latent_encoding = latent_encoding.sample()

        # Figure out the standard deviation of the latent space using
        # the first batch of the data. This is to scale it to have unit variance
        # stabilizes denoising-diffusion training
        if self.global_step == 0 and batch_idx == 0:
            latent_space_scaling_factor = 1 / latent_encoding.std()
            self.latent_space_scaling_factor = latent_space_scaling_factor.float().to(
                self.device
            )

        # Scale the latent encoding to have unit std
        latent_encoding = self.latent_space_scaling_factor * latent_encoding

        loss = self.forward(latent_encoding, labels=sequences)

        self.log("train_loss", loss, prog_bar=True, batch_size=data.size(0))

        return loss

    def validation_step(self, batch: torch.Tensor, batch_idx: int) -> torch.Tensor:
        data, sequences = batch

        with torch.no_grad():
            latent_encoding = self.encoder_model.encode(data)
            latent_encoding = latent_encoding.sample()

        # Scale the latent encoding to have unit std
        latent_encoding = self.latent_space_scaling_factor * latent_encoding

        loss = self.forward(latent_encoding, labels=sequences)

        self.log(
            "epoch_val_loss",
            loss,
            prog_bar=True,
            sync_dist=True,
            batch_size=data.size(0),
        )

        return loss

    @torch.inference_mode()
    def timestep_tests(self, latent_encoding, sequence) -> torch.Tensor:
        # Scale the latent encoding to have unit std
        latent_encoding = self.latent_space_scaling_factor * latent_encoding

        t = torch.arange(0, self.num_timesteps, device=self.device)

        noise = torch.randn_like(latent_encoding)
        x_noised = self.q_sample(latent_encoding, t, noise=noise)

        # nn.Embedding is 0-indexes, so we subtract 1 from the lengths
        lengths = list(map(lambda label: len(label) - 1, [sequence]))

        length_labels = torch.tensor(
            lengths,
            device=latent_encoding.device,
            dtype=torch.long,
        )
        with torch.no_grad():
            length_labels = self.embed_length(length_labels)
            # length_labels = self.length_mlp(length_labels)
            labels = self.sequence2labels([sequence])
            labels = self.mlp(labels)
            labels += length_labels
            predicted_noise = self.model(x_noised, t, labels)[0]

        loss = abs(noise - predicted_noise)
        # loss = loss.mean(dim=0)

        return x_noised

    def configure_optimizers(self):
        """
        Configure the optimizer and the learning rate scheduler for the model.
        Here I am using NVIDIA suggested settings for learning rate and weight
        decay. For ResNet50 they have seen best performance with CosineAnnealingLR,
        initial learning rate of 0.256 for batch size of 256 and linearly scaling
        it down/up for other batch sizes. The weight decay is set to 1/32768 for all
        parameters except the batch normalization layers. For further information check:
        https://catalog.ngc.nvidia.com/orgs/nvidia/resources/resnet_50_v1_5_for_pytorch

        Returns
        -------
        List
            Returns the optimizer and the learning rate scheduler

        Raises
        ------
        ValueError
            If the scheduler is not implemented
        """
        optimizer = torch.optim.AdamW(
            self.model.parameters(),
            lr=self.set_lr,
            betas=(0.9, 0.999),
            eps=1e-08,
            weight_decay=0.01,
            amsgrad=False,
        )

        if self.config_scheduler == "CosineAnnealingWarmRestarts":
            lr_scheduler = {
                "scheduler": CosineAnnealingWarmRestarts(
                    optimizer, T_0=5, eta_min=1e-4
                ),
                "monitor": self.monitor,
                "interval": "epoch",
            }

        elif self.config_scheduler == "OneCycleLR":
            lr_scheduler = {
                "scheduler": OneCycleLR(
                    optimizer,
                    max_lr=0.01,
                    total_steps=self.trainer.estimated_stepping_batches,
                ),
                "monitor": self.monitor,
                "interval": "step",
            }

        elif self.config_scheduler == "CosineAnnealingLR":
            num_epochs = self.trainer.max_epochs
            lr_scheduler = {
                "scheduler": CosineAnnealingLR(
                    optimizer,
                    T_max=num_epochs,
                    eta_min=1e-8,
                ),
                "monitor": self.monitor,
                "interval": "epoch",
            }
        elif self.config_scheduler == "LinearWarmupCosineAnnealingLR":
            num_epochs = self.trainer.max_epochs
            total_steps = self.trainer.estimated_stepping_batches
            steps_per_epoch = total_steps // num_epochs
            # Warmup for 5% of the total steps
            warmup_steps = steps_per_epoch * int(num_epochs * 0.05)

            def lr_lambda(current_step):
                if current_step < warmup_steps:
                    # Linear warmup phase
                    return current_step / max(1, warmup_steps)
                else:
                    # Cosine annealing phase
                    eta_min = 1e-8
                    remaining_steps = current_step - warmup_steps
                    current_epoch = remaining_steps // steps_per_epoch
                    cosine_factor = 0.5 * (
                        1 + math.cos(math.pi * current_epoch / num_epochs)
                    )
                    return eta_min + (1 - eta_min) * cosine_factor

            lr_scheduler = {
                "scheduler": LambdaLR(optimizer, lr_lambda=lr_lambda),
                "monitor": self.monitor,
                "interval": "step",
            }

        else:
            raise ValueError(f"{self.config_scheduler} lr_scheduler is not implemented")

        return [optimizer], [lr_scheduler]


def reduce_sampling_steps(T, K):
    # Calculate the spacing between each sample
    spacing = (T - 1) / (K - 1)

    # Initialize an empty list to store the rounded numbers
    samples = []

    # Generate K evenly spaced real numbers between 1 and T (inclusive)
    for i in range(K):
        # Calculate the current sample
        sample = 1 + spacing * i

        # Round the sample to the nearest integer
        rounded_sample = round(sample)

        # Append the rounded sample to the list
        samples.append(rounded_sample)

    return samples



import os
import time

import mdtraj as md
import numpy as np
import torch
import torch.optim as optim
from scipy.spatial import distance_matrix
from sklearn.manifold import MDS
from tqdm import tqdm

from starling import configs, utilities


def get_tensor_dtype(device):
    """
    Function which returns a tensor dtype based on
    the device passed. This is to support the fact that
    mps does not work with float64 tensors and will set
    them to be cast to float32. This is an internal
    function that really only ends up being relevant
    for gradien descent reconstruction.

    Parameters
    ---------------
    device : torch.device
        Device being used

    returns
    --------------------
    torch.dtype
        Returns the type

    """

    # as of 2024-11 mps does not support float64
    if str(device) == "mps":
        tensor_dtype = torch.float32
    else:
        tensor_dtype = torch.float64

    return tensor_dtype


def compute_pairwise_distances(coords):
    """Function to compute the pairwise distances in 3D space.

    Parameters
    ----------
    coords : torch.Tensor
        A tensor of shape (n, 3) containing the 3D coordinates of n points.

    Returns
    -------
    torch.Tensor
        A tensor of pairwise distances between the points in 3D space.
    """
    return torch.cdist(coords, coords)


def loss_function(original_distance_matrix, coords):
    """
    Function to compute the loss between the original distance
    matrix and the computed distance matrix.

    Parameters
    ----------
    original_distance_matrix : torch.Tensor
        A tensor of shape (n, n) containing the original pairwise
        distances between n points.

    coords : torch.Tensor
        A tensor of shape (n, 3) containing the 3D coordinates
        of n points.

    Returns
    -------
    torch.Tensor
        The mean squared error between the original and computed
        distance matrices, considering only the upper triangle.

    """
    computed_distances = compute_pairwise_distances(coords)

    # Create a mask for the upper triangle
    upper_triangle_mask = torch.triu(
        torch.ones_like(original_distance_matrix), diagonal=1
    ).bool()

    # Apply the mask to both the original and computed distance matrices
    masked_original = original_distance_matrix[upper_triangle_mask]
    masked_computed = computed_distances[upper_triangle_mask]

    # Compute the mean squared error only for the masked elements
    loss = torch.nn.functional.mse_loss(masked_computed, masked_original)

    return loss


def create_incremental_coordinates(n_points, distance, device):
    """
    TO DO: Add docstring
    """

    coordinates = torch.zeros(
        (n_points, 3), dtype=get_tensor_dtype(device), device=device
    )

    # Start the first coordinate at (0, 0, 0)
    for i in range(1, n_points):
        # Generate a random direction vector
        direction = torch.randn(3, dtype=get_tensor_dtype(device), device=device)
        direction /= torch.norm(direction)  # Normalize to get a unit vector

        # Calculate the new coordinate by adding the direction scaled by the distance
        new_coordinate = coordinates[i - 1] + direction * distance

        coordinates[i] = new_coordinate

    return torch.nn.Parameter(coordinates)


def distance_matrix_to_3d_structure_torch_mds(
    target_distances,
    batch_size=100,
    n_iter=300,
    tol=1e-4,
    device="cuda",
    progress_bar=True,
):
    """
    SMACOF implementation using PyTorch with support for batched processing.

    NB; as of Feb 2024 this is substantially slower for MPS than the other approach,
    because MPS seems to fail and fall back on CPU in this parallelized version. This
    is 1.5-2x slower than the other approach. To keep this simple, there is a check in
    the parent function generate_3d_coordinates_from_distances() that defaults to the
    non-torch implementation if the device is MPS and we're on macOS <= 14. Hopefully
    this is fixed in macOS15...

    Parameters:
    -----------
    target_distances: pytorch.Tensor
        tensor of shape (total_samples, n_points, n_points)

    batch_size: int
        size of each processing batch.

    n_iter: int
        Maximum number of iterations.

    tol: float
        Convergence tolerance.

    device: str
        Device to use for computation.

    progress_bar: bool
        Whether to display a progress bar.

    Returns:
    --------
    tuple

        [0] pytorch.Tensor; A tensor of shape (total_samples, n_points, 3)
        [1] pytorch.Tensor; A tensor of shape (total_samples, n_iter)

    """
    # if we passed in a numpy array, convert it to a torch tensor
    # on the correct device
    if isinstance(target_distances, np.ndarray):
        target_distances = torch.from_numpy(target_distances).to(device=device)

    total_samples = target_distances.shape[0]
    n_points = target_distances.shape[1]
    dim = 3
    eps = 1e-12

    X_results = []
    stress_results = []

    # start the progress bar if requested
    if progress_bar == True:
        local_generator = tqdm(range(0, total_samples, batch_size))
    else:
        local_generator = range(0, total_samples, batch_size)

    # for start in range(0, total_samples, batch_size):
    for start in local_generator:
        end = min(start + batch_size, total_samples)
        batch_distances = target_distances[start:end].to(device)

        # Initialize random coordinates for the current batch_size
        X = torch.randn(end - start, n_points, dim, dtype=torch.float32, device=device)
        X = X - X.mean(dim=1, keepdim=True)

        # Initialize stress tracking
        stress_history = torch.zeros(
            end - start, n_iter, dtype=torch.float32, device=device
        )
        old_stress = torch.full(
            (end - start,), float("inf"), dtype=torch.float32, device=device
        )
        converged = torch.zeros(end - start, dtype=torch.bool, device=device)

        for it in range(n_iter):
            diff = X.unsqueeze(2) - X.unsqueeze(1)
            D = torch.norm(diff, dim=3) + eps

            stress = torch.sum((D - batch_distances) ** 2, dim=(1, 2))
            stress_history[:, it] = stress

            converged = converged | (torch.abs(old_stress - stress) < tol)
            if torch.all(converged):
                stress_history = stress_history[:, : it + 1]
                break

            B = torch.zeros_like(D)
            mask = D > eps

            # This line is equivalent to
            # B[mask] = -batch_distances[mask] / D[mask]
            # but much faster and no issues across OS's [bugs on some macOS's (<= 14)]
            B = torch.where(mask, -batch_distances / D, B)

            row_sums = B.sum(dim=2)
            B.diagonal(dim1=1, dim2=2).copy_(-row_sums)

            X_new = torch.bmm(B, X) / n_points

            # X[~converged] = X_new[~converged]
            # X[~converged] = X[~converged] - X[~converged].mean(dim=1, keepdim=True)
            # switching the above to the torch.where equivalents
            X = torch.where(converged.unsqueeze(1).unsqueeze(2), X, X_new)
            X = torch.where(
                converged.unsqueeze(1).unsqueeze(2), X, X - X.mean(dim=1, keepdim=True)
            )

            old_stress = stress

        X_results.append(X.cpu())
        stress_results.append(stress_history.cpu())

    # close the progress bar if opened
    if progress_bar == True:
        local_generator.close()

    # Concatenate all chunk results
    X_final = torch.cat(X_results, dim=0)
    stress_final = torch.cat(stress_results, dim=0)

    return X_final.numpy(), stress_final.numpy()


def distance_matrix_to_3d_structure_mds(distance_matrix, **kwargs):
    """
    Generate 3D coordinates from a distance matrix using
    multidimensional scaling (MDS).

    NB: by default the MDS object is initialized with
    the following defaults, although these can be overridden
    by passing them in the kwargs:

    n_components = 3
    dissimilarity = "precomputed"
    n_init = configs.DEFAULT_MDS_NUM_INIT  (default setting)
    n_jobs = configs.DEFAULT_CPU_COUNT_MDS (default setting)
    normalized_stress = 'auto'

    Parameters
    ----------
    distance_matrix : torch.Tensor
        A 2D tensor representing the distance matrix.

    kwargs : dict
        Keyword arguments to pass to scikit-learn's MDS
        algorithm.

    Returns
    -------
    torch.Tensor
        A 3D tensor representing the coordinates of the
        atoms.

    """
    # Set the default values for n_init and n_jobs if not provided in kwargs
    # this matches the default values in scikit-learn's MDS
    n_init = kwargs.pop("n_init", configs.DEFAULT_MDS_NUM_INIT)
    n_jobs = kwargs.pop("n_jobs", configs.DEFAULT_CPU_COUNT_MDS)

    # Initialize MDS with 3 components (for 3D) and the specified parameters
    # nb: normalized_stress = 'auto' explicitly as this is the default
    # value in sci-kit learn >1.4 , but before that it was False, so this just
    # ensures version-independent behavior in the MDS call
    mds = MDS(
        n_components=3,
        dissimilarity="precomputed",
        n_init=n_init,
        n_jobs=n_jobs,
        normalized_stress="auto",
        **kwargs,
    )

    # Fit the MDS model to the distance matrix
    coords = mds.fit_transform(distance_matrix)

    return coords


def distance_matrix_to_3d_structure_gd(
    original_distance_matrix,
    num_iterations=5000,
    learning_rate=1e-3,
    device="cuda:0",
    verbose=True,
):
    """
    Function to reconstruct a 3D structure from a
    distance matrix using gradient descent.

    Parameters
    ----------
    original_distance_matrix : torch.Tensor or numpy.ndarray
        The original distance matrix.

    num_iterations : int, optional
        Number of iterations for gradient descent, by default 5000.

    learning_rate : float, optional
        Learning rate for the optimizer, by default 1e-3.

    device : str, optional
        Device to which tensors are moved, by default "cuda:0".

    verbose : bool, optional
        Whether to print progress, by default True.

    Returns
    -------
    numpy.ndarray
        The reconstructed 3D coordinates.

    NB: As of Nov 2024, Apple MPS does not support float64 tensors, so we cast
    tensors to float32 in the case of MPS being used. Note that this is actually
    slower than using CPU, but we provide support for this in case someone wants
    it and/or it gets faster in the future...

    """

    if isinstance(original_distance_matrix, torch.Tensor):
        original_distance_matrix = original_distance_matrix.to(
            device, dtype=get_tensor_dtype(device)
        )

    else:
        original_distance_matrix = torch.tensor(
            original_distance_matrix, dtype=get_tensor_dtype(device), device=device
        )

    coords = create_incremental_coordinates(
        original_distance_matrix.shape[0], 3.6, device=device
    )
    # coords = torch.randn((original_distance_matrix.size(0), 3), requires_grad=True, device=device,dtype=torch.float64)

    # optimizer = optim.SGD([coords], lr=learning_rate, momentum=0.99, nesterov=True)
    optimizer = optim.Adam([coords], lr=learning_rate)

    for i in range(num_iterations):
        optimizer.zero_grad()

        loss = loss_function(original_distance_matrix, coords)

        loss.backward()

        optimizer.step()

        if i % 100 == 0 and verbose:
            print(f"Iteration {i}, Loss: {loss.item()}")

    return coords.detach().cpu().numpy()


def compare_distance_matrices(original_distance_matrix, coords, return_abs_diff=True):
    """
    Function to compare the original distance matrix with the
    computed distance matrix.

    Parameters
    ----------
    original_distance_matrix : np.ndarray
        The original distance matrix.

    coords : np.ndarray
        The computed 3D coordinates.

    return_abs_diff : bool
        Whether to return the absolute difference between the
        original and computed distance matrices, or the signed
        difference (original - computed), by default True.

    Returns
    -------
    tuple (np.ndarray, np.ndarray)
        [0] - The distance matrix computed from the 3D coordinates.
        [1] - The absolute difference between the original and computed
              distance matrices.
    """

    # compute the redundant inter-residue distance map based on the
    # passed coordinates
    computed_distance_matrix = distance_matrix(coords, coords)

    # calculate the difference between the original distance map and the distance map
    # derived from the input 3D structure
    difference_matrix = original_distance_matrix - computed_distance_matrix

    if return_abs_diff:
        difference_matrix = np.abs(difference_matrix)

    # return the computed distance matrix and the difference matrix
    return computed_distance_matrix, difference_matrix


def create_ca_topology_from_coords(sequence, coords):
    """
    Creates a CA backbone topology from a protein sequence and 3D coordinates.

    Parameters
    ----------
    sequence : str
        Protein sequence as a string of amino acid single-letter codes.

    coords : np.ndarray
        3D coordinates for each CA atom.

    Returns
    ----------
    md.Trajectory
         MDTraj trajectory object containing the CA backbone topology and coordinates.
    """

    # Create an empty topology
    topology = md.Topology()

    # Add a chain to the topology
    chain = topology.add_chain()

    # -- topology construction loop

    for i, res in enumerate(sequence):
        try:
            res_three_letter = configs.AA_ONE_TO_THREE[res]
        except KeyError:
            raise ValueError(f"Invalid amino acid: {res}")

        residue = topology.add_residue(res_three_letter, chain)

        # Add a CA atom to the residue
        ca_atom = topology.add_atom("CA", md.element.carbon, residue)

        # Connect the CA atom to the previous CA atom (if not the first residue)
        if i > 0:
            topology.add_bond(topology.atom(i - 1), ca_atom)

    # --- end of topology construction loop

    # Ensure the coordinates are in the right shape (1, num_atoms, 3)
    if coords.ndim != 3:
        coords = coords[np.newaxis, :, :]

    # commented out for now
    # else:
    #    print(coords.shape)

    # Create an MDTraj trajectory object with the topology and coordinates
    traj = md.Trajectory(coords, topology)

    return traj


# Function to save the MDTraj trajectory to a specified file
def save_trajectory(traj, filename):
    """
    Saves the MDTraj trajectory to a specified file. This invokes
    the trak.save() method of the MDTraj trajectory object.

    Parameters
    -----------
    traj : md.Trajectory
        The MDTraj trajectory object to save.

    filename : str
        The name of the file to save the trajectory .
    """

    traj.save(filename)


def generate_3d_coordinates_from_distances(
    device, batch_size, num_cpus_mds, num_mds_init, distance_maps, progress_bar=True
):
    """
    Function to generate 3D coordinates from distance maps.

    This is the parent function which will then choose either
    the fasts or most approriate method to use based on the
    requested device and the actual OS.

    This function is called in:

        1. inference.generation, where the distance maps are
           tensors.

        2. From the Ensemble() object, where the distance maps
           are numpy arrays.

    Parameters
    ----------
    device : str
        The device to use for computation.

    batch_size : int
        The batch size for processing the distance maps if
        we use the torch implementation.

    num_cpus_mds : int
        The number of CPUs to use for MDS if we use the
        SciPy implementation.

    num_mds_init : int
        The number of initializations to use for MDS if
        we use the SciPy implementation.

    distance_maps : list
        Either an nd.array or a pytorch tensor of distance
        maps.
    """

    # cast device to string for ffs
    device = str(device)

    # leaving this in case we have to revert
    # note; macOS <= 14  gives a "MPS: nonzero op" error, so because of
    # this we force 'cpu' here if on macOS
    # macOS_version = utilities.get_macOS_version()

    # if (macOS_version > 0 and macOS_version < 15) or device == "cpu":
    if device == "cpu":
        ## NB: It seemed like we needed to do this at one point, but
        ## maybe not anymore? Leaving here in case we need it later
        
        # if we passed in a numpy array, convert it to a torch tensor
        # on the correct device (note if we're here we're on CPU)
        # if isinstance(distance_maps, np.ndarray):
        #    distance_maps = torch.from_numpy(distance_maps).to('cpu')

        # open a progress bar if requested
        # Open a progress bar if requested
        dm_generator = tqdm(distance_maps, total=len(distance_maps)) if progress_bar else distance_maps

        # loop over each distance map
        coordinates = []
        for dist_map in dm_generator:
            coord = distance_matrix_to_3d_structure_mds(dist_map, n_jobs=num_cpus_mds, n_init=num_mds_init)            
            coordinates.append(coord)

        # cast to array and convert to nm
        coordinates = np.array(coordinates) / configs.CONVERT_ANGSTROM_TO_NM
    else:
        # call the torch MDS implementation
        coordinates, _ = distance_matrix_to_3d_structure_torch_mds(
            distance_maps,
            batch_size=batch_size,
            device=device,
            progress_bar=progress_bar,
        )
        coordinates /= configs.CONVERT_ANGSTROM_TO_NM

    return coordinates



import torch


def linear_beta_schedule(timesteps: int) -> torch.Tensor:
    """
    The beta values are linearly spaced between 0.0001 and 0.02.

    Parameters
    ----------
    timesteps : int
        The number of timesteps which will be used to generate the beta values.

    Returns
    -------
    torch.Tensor
        A tensor containing the beta values.
    """
    scale = 1000 / timesteps
    beta_start = scale * 0.0001
    beta_end = scale * 0.02
    return torch.linspace(beta_start, beta_end, timesteps, dtype=torch.float64)


def cosine_beta_schedule(timesteps: int, s: float = 0.008) -> torch.Tensor:
    """
    The beta values are generated using a cosine function. The beta values are
    clipped between 0.0001 and 0.9999.

    Parameters
    ----------
    timesteps : int
        The number of timesteps which will be used to generate the beta values.
    s : float, optional
        Adjusts the smoothness of the beta schedule's initial portion, influencing
        how quickly the values change over the first few timesteps. By default 0.008.

    Returns
    -------
    torch.Tensor
        A tensor containing the beta values.
    """
    steps = timesteps + 1
    x = torch.linspace(0, timesteps, steps)
    alphas_cumprod = torch.cos(((x / timesteps) + s) / (1 + s) * torch.pi * 0.5) ** 2
    alphas_cumprod = alphas_cumprod / alphas_cumprod[0]
    betas = 1 - (alphas_cumprod[1:] / alphas_cumprod[:-1])
    return torch.clip(betas, 0.0001, 0.9999)


def sigmoid_beta_schedule(
    timesteps: int, start: int = 3, end: int = 3, tau: int = 1
) -> torch.Tensor:
    """
    The beta values are generated using a sigmoid function. The beta values are
    clipped between 0 and 0.999.

    Parameters
    ----------
    timesteps : int
        The number of timesteps which will be used to generate the beta values.
    start : int, optional
        The starting value for the sigmoid function, by default 3
    end : int, optional
        The ending value for the sigmoid function, by default 3
    tau : int, optional
        The time constant for the sigmoid function, by default 1

    Returns
    -------
    torch.Tensor
        A tensor containing the beta values.
    """
    steps = timesteps + 1
    t = torch.linspace(0, timesteps, steps, dtype=torch.float64) / timesteps
    v_start = torch.tensor(start / tau).sigmoid()
    v_end = torch.tensor(end / tau).sigmoid()
    alphas_cumprod = (-((t * (end - start) + start) / tau).sigmoid() + v_end) / (
        v_end - v_start
    )
    alphas_cumprod = alphas_cumprod / alphas_cumprod[0]
    betas = 1 - (alphas_cumprod[1:] / alphas_cumprod[:-1])
    return torch.clip(betas, 0, 0.999)



from typing import Tuple

import numpy as np
import torch
from torch import nn
from tqdm import tqdm

from starling.data.data_wrangler import one_hot_encode


class DDIMSampler(nn.Module):
    def __init__(
        self,
        ddpm_model,
        n_steps: int,
        ddim_discretize: str = "uniform",
        ddim_eta: float = 0.0,
    ):
        """
        An efficient sampler that can output samples 10x to 100x faster than the DDPM model.
        Denosing diffusion implicit models (DDIM) do not require to sample the entire diffusion process to generate samples.
        The forward process is parameterized using non-Markovian diffusion processes, which leads to short generative Markov chains
        that can be simulated in a small number of steps.


        Parameters
        ----------
        ddpm_model : _type_
            The trained DDPM model.
        n_steps : int
            The number of steps to simulate the generative process. Smaller than the number of steps used to train the DDPM model.
        ddim_discretize : str, optional
            The discretization method to use for the generative process, by default "uniform"
        ddim_eta : float, optional
            The noise level for the generative process, a number between 0.0 and 1.0.
            0.0 adds no noise to the generative process, 1.0 adds the maximum noise.
            This number interpolates between deterministic and stochastic generative processes,
            by default 0.0

        Raises
        ------
        NotImplementedError
            If the discretization method is not implemented.
        """
        super(DDIMSampler, self).__init__()
        self.ddpm_model = ddpm_model
        self.n_steps = self.ddpm_model.num_timesteps
        self.ddim_discretize = ddim_discretize
        self.ddim_eta = ddim_eta

        # Ways to discretize the generative process
        if ddim_discretize == "uniform":
            c = self.n_steps // n_steps
            self.ddim_time_steps = np.asarray(list(range(0, self.n_steps - 1, c))) + 1
        elif ddim_discretize == "quad":
            self.ddim_time_steps = (
                (np.linspace(0, np.sqrt(self.n_steps * 0.8), n_steps)) ** 2
            ).astype(int) + 1
        else:
            raise NotImplementedError(ddim_discretize)

        with torch.no_grad():
            alpha_bar = self.ddpm_model.alphas_cumprod
            self.ddim_alpha = alpha_bar[self.ddim_time_steps].clone().to(torch.float32)
            self.ddim_alpha_sqrt = torch.sqrt(self.ddim_alpha)
            self.ddim_alpha_prev = torch.cat(
                [alpha_bar[0:1], alpha_bar[self.ddim_time_steps[:-1]]]
            )
            self.ddim_sigma = (
                ddim_eta
                * (
                    (1 - self.ddim_alpha_prev)
                    / (1 - self.ddim_alpha)
                    * (1 - self.ddim_alpha / self.ddim_alpha_prev)
                )
                ** 0.5
            )

            self.ddim_sqrt_one_minus_alpha = (1.0 - self.ddim_alpha) ** 0.5

    def generate_labels(self, labels: str) -> torch.Tensor:
        """
        Generate labels to condition the generative process on.

        Parameters
        ----------
        labels : str
            A sequence to generate labels from.

        Returns
        -------
        torch.Tensor
            The labels to condition the generative process on.
        """
        labels = (
            torch.argmax(
                torch.from_numpy(one_hot_encode(labels.ljust(384, "0"))), dim=-1
            )
            .to(torch.int64)
            .squeeze()
            .to(self.ddpm_model.device)
        )

        labels = self.ddpm_model.sequence2labels(labels)

        labels = labels.unsqueeze(0)

        return labels

    @torch.no_grad()
    def sample(
        self,
        num_conformations: int,
        labels: torch.Tensor,
        repeat_noise: bool = False,
        temperature: float = 1.0,
        show_per_step_progress_bar: bool = True,
    ) -> torch.Tensor:
        """
        Sample the generative process using the DDIM model.

        Parameters
        ----------
        num_conformations : int
            Number of conformations to generate.
        labels : torch.Tensor
            The labels to condition the generative process on.
        repeat_noise : bool, optional
            _description_, by default False
        temperature : float, optional
            _description_, by default 1.0
        show_per_step_progress_bar : bool, optional
            whether to show progress bar per step.

        Returns
        -------
        torch.Tensor
            The generated distance maps.
        """
        device = self.ddpm_model.device

        # Initialize the latents with noise
        x = torch.randn(
            [
                num_conformations,
                self.ddpm_model.in_channels,
                self.ddpm_model.image_size,
                self.ddpm_model.image_size,
            ],
            device=device,
        )

        time_steps = np.flip(self.ddim_time_steps)

        # Get the labels to condition the generative process on
        labels = self.generate_labels(labels)

        # initialize progress bar if we want to show it
        if show_per_step_progress_bar:
            pbar = tqdm(len(time_steps))

        # Denoise the initial latent
        for i, step in enumerate(time_steps):
            index = len(time_steps) - i - 1

            # Batch the timesteps
            ts = x.new_full((num_conformations,), step, dtype=torch.long)

            # Sample the generative process
            x, *_ = self.p_sample(
                x,
                labels,
                ts,
                step,
                index=index,
                repeat_noise=repeat_noise,
                temperature=temperature,
            )
            # update progress bar if we are showing it
            if show_per_step_progress_bar:
                pbar.update(1)

        # if we have progress bar, close after finishing the steps.
        if show_per_step_progress_bar:
            pbar.close()

        # Scale the latents back to the original scale
        x = (1 / self.ddpm_model.latent_space_scaling_factor) * x

        # Decode the latents to get the distance maps
        x = self.ddpm_model.encoder_model.decode(x)

        return x

    @torch.no_grad()
    def p_sample(
        self,
        x: torch.Tensor,
        c: torch.Tensor,
        t: torch.Tensor,
        step: int,
        index: int,
        repeat_noise: bool = False,
        temperature: float = 1.0,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Take one step in the generative process.

        Parameters
        ----------
        x : torch.Tensor
            The tensor to remove noise from.
        c : torch.Tensor
            The labels to condition the generative process on.
        t : torch.Tensor
            The timestep to sample the generative process at.
        step : int

        index : int
            _description_
        repeat_noise : bool, optional
            _description_, by default False
        temperature : float, optional
            _description_, by default 1.0

        Returns
        -------
        _type_
            _description_
        """

        # Predict the amount of noise in the latent based on the timestep and labels

        predicted_noise = self.ddpm_model.model(x, t, c)

        # Calculate the previous latent and the predicted latent
        x_prev, pred_x0 = self.get_x_prev_and_pred_x0(
            predicted_noise,
            index,
            x,
            temperature=temperature,
            repeat_noise=repeat_noise,
        )

        return x_prev, pred_x0, predicted_noise

    def get_x_prev_and_pred_x0(
        self,
        predicted_noise: torch.Tensor,
        index: int,
        x: torch.Tensor,
        temperature: float,
        repeat_noise: bool,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Remove the noise from the latent iteratively.

        Parameters
        ----------
        predicted_noise : torch.Tensor
            The noise predicted by the DDPM model.
        index : int
            The index of the timestep.
        x : torch.Tensor
            The latent to remove the noise from.
        temperature : float
            The temperature to use for the generative process
        repeat_noise : bool
            Whether to repeat the noise

        Returns
        -------
        Tuple[torch.Tensor, torch.Tensor]
            The previous latent.
        """
        alpha = self.ddim_alpha[index]

        alpha_prev = self.ddim_alpha_prev[index]

        sigma = self.ddim_sigma[index]

        sqrt_one_minus_alpha = self.ddim_sqrt_one_minus_alpha[index]

        # Predicted x_0
        pred_x0 = (x - sqrt_one_minus_alpha * predicted_noise) / (alpha**0.5)

        # Direction pointing to x_t
        dir_xt = (1.0 - alpha_prev - sigma**2).sqrt() * predicted_noise

        if sigma == 0.0:
            noise = 0.0

        elif repeat_noise:
            noise = torch.randn((1, *x.shape[1:]), device=x.device)
        else:
            noise = torch.randn(x.shape, device=x.device)

        noise = noise * temperature

        x_prev = (alpha_prev**0.5) * pred_x0 + dir_xt + sigma * noise

        return x_prev, pred_x0



#
# Core utilities for the package. This should not importa anything from within starling
# to avoid circular imports.
#

import gzip
import lzma
import os
import pickle
import warnings
import platform

import numpy as np
import torch

# code that allows access to the data directory
_ROOT = os.path.abspath(os.path.dirname(__file__))


def get_data(path):
    return os.path.join(_ROOT, "data", path)


def fix_ref_to_home(input_path):
    """
    Function to fix the path to the home directory.

    Parameters
    ---------------
    path : str
        The path to fix.

    Returns
    ---------------
    str: The fixed path.
    """
    if input_path.startswith("~"):
        return os.path.expanduser(input_path)
    return input_path


def check_file_exists(input_path):
    """
    Function to check if a file exists.

    Parameters
    ---------------
    path : str
        The path to check.

    Returns
    ---------------
    bool: True if the file exists, False otherwise.
    """
    return os.path.exists(input_path) and os.path.isfile(input_path)


def remove_extension(input_path, verbose=True):
    """
    Function to remove the extension from a file.

    Parameters
    ---------------
    path : str
        The path to remove the extension from.

    verbose : str
        Define how noisy to be...

    Returns
    ---------------
    str
        The path with the extension removed.

    """
    new_filename = os.path.splitext(input_path)[0]

    # added this in so we don't silently edit away a filename with a period
    # that would be invisible...
    if verbose:
        if input_path != new_filename:
            print(
                f"Warning: removed file extension from input file name.\nWas: {input_path}\nNow: {new_filename}"
            )

    return new_filename


def parse_output_path(args):
    """
    Parse the output path from the command line arguments.

    Parameters
    ----------
    args : argparse.Namespace
        The command line arguments.

    Returns
    -------
    str
        The output path and filename without an extension.
    """

    # get the filename (+extension) if the input file, without
    # any path info
    input_filename = os.path.basename(args.input_file)

    # if no output is specified, use the current directory;
    # this is the default behavior
    if args.output == ".":
        outname = input_filename

    # if input was provided for the output
    else:
        # if we were passed a path
        if os.path.isdir(args.output):
            outname = os.path.join(args.output, input_filename)
        else:
            outname = args.output
    # remove the extension
    outname = os.path.splitext(outname)[0]

    return outname

def get_macOS_version():
    """
    Function to check the macOS version.

    Returns
    ---------------
    int
        Returns -1 or the major macOS version.
    
    """
    # get the macOS version
    macos_version = platform.mac_ver()[0]

    if len(macos_version) == 0:
        return -1
    else:
        try:
            return int(macos_version.split('.')[0])
        except ValueError:
            return -1



def check_device(use_device, default_device="gpu"):
    """
    Function to check the device was correctly set.

    Parameters
    ---------------
    use_device : str
        Identifier for the device to be used for predictions.
        Possible inputs: 'cpu', 'mps', 'cuda', 'cuda:int', where the int corresponds to
        the index of a specific cuda-enabled GPU. If 'cuda' is specified and
        cuda.is_available() returns False, this will raise an Exception
        If 'mps' is specified and mps is not available, an exception will be raised.

    default_device : str
        The default device to use if device=None.
        If device=None and default_device != 'cpu' and default_device is
        not available, device_string will be returned as 'cpu'.
        Default is 'gpu'. This checks first for cuda and then for mps
        because STARLING is faster on both than it is on CPU, so we should
        use the fastest device available.
        Options are 'cpu' or 'gpu'

    Returns
    ---------------
    torch.device: A PyTorch device object representing the device to use.
    """

    # Helper function to get CUDA device string (e.g., 'cuda:0', 'cuda:1')
    def get_cuda_device(cuda_str):
        if cuda_str == "cuda":
            return torch.device("cuda")
        else:
            device_index = int(cuda_str.split(":")[1])
            num_devices = torch.cuda.device_count()
            if device_index >= num_devices:
                raise ValueError(
                    f"{cuda_str} specified, but only {num_devices} CUDA-enabled GPUs are available. "
                    f"Valid device indices are from 0 to {num_devices - 1}."
                )
            return torch.device(cuda_str)

    # If `use_device` is None, fall back to `default_device`
    if use_device is None:
        default_device = default_device.lower()
        if default_device == "cpu":
            return torch.device("cpu")
        elif default_device == "gpu":
            if torch.cuda.is_available():
                return torch.device("cuda")
            elif torch.backends.mps.is_available():
                return torch.device("mps")
            else:
                return torch.device("cpu")
        else:
            raise ValueError("Default device can only be set to 'cpu' or 'gpu'.")

    # if a device is passed as torch.device, change to string so we
    # can make lowercase str for handling different device types.
    if isinstance(use_device, torch.device):
        use_device = str(use_device)

    # Ensure `use_device` is a string
    if not isinstance(use_device, str):
        raise ValueError(
            "Device must be type torch.device or string, valid options are: 'cpu', 'mps', 'cuda', or 'cuda:int'."
        )

    # make lower case to make checks easier.
    use_device = use_device.lower()

    # Handle specific device strings
    if use_device == "cpu":
        return torch.device("cpu")

    elif use_device == "mps":
        if torch.backends.mps.is_available():
            return torch.device("mps")
        else:
            raise ValueError(
                "MPS was specified, but MPS is not available. Make sure you're running on an Apple device with MPS support."
            )

    elif use_device.startswith("cuda"):
        if not torch.cuda.is_available():
            raise ValueError(
                f"{use_device} was specified, but torch.cuda.is_available() returned False."
            )
        return get_cuda_device(use_device)

    else:
        raise ValueError(
            "Device must be 'cpu', 'mps', 'cuda', or 'cuda:int' (where int is a valid GPU index)."
        )

    # This should never be reached
    raise RuntimeError(
        "Unexpected state in the check_device function. Please raise an issue on GitHub."
    )


def write_starling_ensemble(
    ensemble_object,
    filename,
    compress=False,
    reduce_precision=None,
    compression_algorithm="lzma",
    verbose=True,
):
    """
    Function to write the STARLING ensemble to a file in the STARLING
    format (.starling). This is actially just a dictionary with the
    amino acid sequence, the distance maps, and the SSProtein object
    if available.

    Parameters
    ---------------
    ensemble_object : starling.structure.Ensemble
        The STARLING ensemble object to save to a file.

    filename : str
        The filename to save the ensemble to; note this should
        not include a file extenison and if it does this will
        be removed

    compress : bool
        Whether to compress the file or not. Default is False.

    reduce_precision : bool
        Whether to reduce the precision of the distance map to a
        single decimal point and cast to float16 if possible.
        Default is None. Sets to False if compression is False,
        but True if compression is True.

    compression_algorithm : str
        The compression algorithm to use. Options are 'gzip' and 'lzma'.
        `lzma` gives better compression if reduce_precision is set to True,
        but actually 'gzip' is better if reduce_precision is False. 'lzma'
        is also slower than 'gzip'. Default is 'lzma'.

    verbose : bool
        Flag to define how noisy we should be
    """

    # set reduce_precision to mirror compress if not set
    if reduce_precision is None:
        reduce_precision = compress

    # build_the save dictionary
    save_dict = {
        "sequence": ensemble_object.sequence,
        "distance_maps": ensemble_object._Ensemble__distance_maps,
        "traj": ensemble_object._Ensemble__trajectory,
        "DEFAULT_ENCODER_WEIGHTS_PATH": ensemble_object._Ensemble__metadata[
            "DEFAULT_ENCODER_WEIGHTS_PATH"
        ],
        "DEFAULT_DDPM_WEIGHTS_PATH": ensemble_object._Ensemble__metadata[
            "DEFAULT_DDPM_WEIGHTS_PATH"
        ],
        "VERSION": ensemble_object._Ensemble__metadata["VERSION"],
        "DATE": ensemble_object._Ensemble__metadata["DATE"],
    }

    # if we wish to reduce the precision of the distance map to a single decimal point
    if reduce_precision:
        # run here so we catch warnings if they happen
        with warnings.catch_warnings(record=True) as w:
            # cast to float16 array and round to 1 decimal place
            tmp = np.round(ensemble_object._Ensemble__distance_maps, decimals=1).astype(
                "float16"
            )

            # check if a RuntimeWarning was raised
            for warning in w:
                # IF yes, then we actually don't cast because we can't do so faithfully
                if issubclass(warning.category, RuntimeWarning):
                    print(
                        "Warning: Could not reduce precision of distance maps to float16. Saving as float64."
                    )
                    tmp = np.round(ensemble_object._Ensemble__distance_maps, decimals=1)

        # update the save dictionary one way or another.
        save_dict["distance_maps"] = tmp

    # Remove the extension if it exists
    filename = remove_extension(filename, verbose=verbose)

    # add starling extension
    filename = filename + ".starling"

    # If we want to compress the file
    if compress:
        if compression_algorithm == "gzip":
            with gzip.open(filename + ".gzip", "wb") as file:
                pickle.dump(save_dict, file)

        elif compression_algorithm == "lzma":
            with lzma.open(filename + ".xz", "wb") as file:
                pickle.dump(save_dict, file)
        else:
            raise ValueError(
                f"Compression algorithm {compression_algorithm} is not supported. Supported algorithms are 'gzip' and 'lzma'."
            )

    else:
        with open(filename, "wb") as file:
            pickle.dump(save_dict, file)


def read_starling_ensemble(filename):
    """
    Function to read a STARLING ensemble from a file in the STARLING
    format (.starling) or a compressed starling file (.gzip, .xz).
    The .starling file is actially just a dictionary with the
    amino acid sequence, the distance maps, some metadata and the
    SSProtein object if available.

    Note this determines the file type based on the extension and
    will raise an error if the file is not a .starling file, a
    a .starling.gzip or .starling.xz file.

    Parameters
    ---------------
    filename : str
        The filename to read the ensemble from; note this should
        not include a file extenison and if it does this will
        be removed

    Returns
    ---------------
    starling.structure.Ensemble: The STARLING ensemble object
    """

    if filename.endswith(".starling.gzip"):
        try:
            with gzip.open(filename, "rb") as file:
                return_dict = pickle.load(file)
        except Exception:
            raise ValueError(
                f"Could not read the file {filename}. Please check the path and try again."
            )

    elif filename.endswith(".starling.xz"):
        try:
            with lzma.open(filename, "rb") as file:
                return_dict = pickle.load(file)
        except Exception:
            raise ValueError(
                f"Could not read the file {filename}. Please check the path and try again."
            )

    elif filename.endswith(".starling"):
        try:
            with open(filename, "rb") as file:
                return_dict = pickle.load(file)
        except Exception:
            raise ValueError(
                f"Could not read the file {filename}. Please check the path and try again."
            )

    else:
        raise ValueError(
            f"File {filename} does not have the extension of .starling.gzip, .starling.xz or .starling."
        )

    # NOTE: We if the distance maps are float16, we cast them to float32
    if return_dict["distance_maps"].dtype == "float16":
        return_dict["distance_maps"] = return_dict["distance_maps"].astype("float32")

    return return_dict


def symmetrize_distance_maps(dist_maps):
    """
    Symmetrizes a stack of distance maps along an axis by reflecting the upper triangle
    onto the lower triangle and setting the diagonal values to zero.

    Parameters:
    ----------
    dist_maps : np.ndarray
        A 3D NumPy array of shape (N, M, M), where N is the number of distance maps.

    Returns:
    -------
    np.ndarray
        A symmetrized stack of distance maps.
    """
    # Ensure input is a 3D array
    assert dist_maps.ndim == 3, "Input must be a 3D array of shape (N, M, M)."
    assert dist_maps.shape[1] == dist_maps.shape[2], "Each distance map must be square."

    # Create masks for the upper triangle
    M = dist_maps.shape[1]
    mask_upper_triangle = np.triu_indices(M, k=1)

    # Reflect the upper triangle onto the lower triangle
    dist_maps[:, mask_upper_triangle[1], mask_upper_triangle[0]] = dist_maps[
        :, mask_upper_triangle[0], mask_upper_triangle[1]
    ]

    # Set diagonal values to zero
    np.einsum("nii->ni", dist_maps)[:] = 0  # Efficient diagonal zeroing

    return dist_maps


def get_off_diagonals(
    distance_map, min_separation=1, max_separation=4, return_mean=False
):
    """
    Function to calculate the  the off-diagonal elements of a matrix.

    This is useful for error checking as we can KNOW the max distance between
    any pair of residues base on the contour length of the protein, allowing
    us to identify conformations that are simply impossible.

    Parameters
    ---------------
    distance_map : np.ndarray
        The distance map to check for errors.

    min_separation : int
        The minimum sequence separation to check across.

    max_separation : int
        The maxiumum sequence separation to check across.

    Returns
    ---------------
    np.ndarray
        All off-diagonal elements of the distance map from
        min_separation to max_separation away from the true
        diagonal.

    """

    # check we have a square...
    if distance_map.shape[0] != distance_map.shape[1]:
        raise ValueError("Input matrix must be square.")

    # get sequence length
    n = distance_map.shape[0]

    # build a single list with all the diagonal vals
    values = []
    for d in range(min_separation, max_separation + 1):
        values.extend(distance_map.diagonal(d))  # Upper diagonals
        values.extend(distance_map.diagonal(-d))  # Lower diagonals

    if return_mean:
        return np.mean(values)
    else:
        return np.array(values)


def check_distance_map_for_error(distance_map, min_separation=1, max_separation=2):
    """
    Function to check the distance map for errors.

    Parameters
    ---------------
    distance_map : np.ndarray
        The distance map to check for errors.

    min_separation : int
        The minimum sequence separation to check across.

    max_separation : int
        The maxiumum sequence separation to check across.

    Returns
    ---------------
    bool
        Returns True if an error was detected,
        and False if not.

    """

    ij_abs = abs(max_separation - min_separation) + 1

    # we can KNOW the max distance any i-j residues are
    # nb 3.81 is Mpipi bond length, but we add a +1 Angstrom to EACH
    # bond as an error term; this likely makes the efficacy of this approach
    # VERY poor if you have large sequence separation, but it also minimizes
    # the risk of false positives.
    # changes should update this but we'll hardcode it for now...
    max_possible_dist = 4.5 * ij_abs

    # get biggest distance from the set
    max_ods = np.max(
        get_off_diagonals(
            distance_map, min_separation=min_separation, max_separation=max_separation
        )
    )
    if max_ods > max_possible_dist:
        return True
    else:
        return False



import time

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import torch
from IPython import embed
from scipy.spatial.distance import pdist, squareform
from sklearn.manifold import MDS
from tqdm import trange

from starling import configs
from starling.frontend.ensemble_generation import generate
from starling.inference.evaluate_vae import get_errors
from starling.structure.coordinates import (
    create_ca_topology_from_coords,
    distance_matrix_to_3d_structure_gd,
    distance_matrix_to_3d_structure_mds,
    distance_matrix_to_3d_structure_torch_mds,
)
from starling.utilities import get_data


def visualize_comparison(original_dms, scipy_coords, torch_coords):
    """
    Visualization function with color scaling based on min/max errors for both methods.
    """
    fig, axes = plt.subplots(2, 3, figsize=(7, 5))

    # Compute all differences first for consistent scaling
    all_diffs = []
    for i in range(3):
        scipy_dm = squareform(pdist(scipy_coords[i]))
        torch_dm = squareform(pdist(torch_coords[i]))

        scipy_diff = original_dms[i] - scipy_dm
        torch_diff = original_dms[i] - torch_dm

        all_diffs.extend([scipy_diff, torch_diff])

    vmin = min(diff.min() for diff in all_diffs)
    vmax = max(diff.max() for diff in all_diffs)
    abs_max = max(abs(vmin), abs(vmax))

    # Define modern colormap
    cmap = "viridis"

    for i in range(3):
        scipy_dm = squareform(pdist(scipy_coords[i]))
        torch_dm = squareform(pdist(torch_coords[i]))

        scipy_diff = original_dms[i] - scipy_dm
        torch_diff = original_dms[i] - torch_dm

        # Plot with improved aesthetics
        sns.heatmap(
            scipy_diff,
            ax=axes[0, i],
            cmap=cmap,
            center=0,
            vmin=-abs_max,
            vmax=abs_max,
            cbar_kws={"shrink": 0.8},
            square=True,
        )
        axes[0, i].set_title(
            f"Scipy MDS Conformer {i + 1}\nRMSE: {np.sqrt(np.mean(scipy_diff**2)):.2e}",
            pad=10,
            fontsize=9,
        )

        sns.heatmap(
            torch_diff,
            ax=axes[1, i],
            cmap=cmap,
            center=0,
            vmin=-abs_max,
            vmax=abs_max,
            cbar_kws={"shrink": 0.8},
            square=True,
        )
        axes[1, i].set_title(
            f"Torch SMACOF Conformer {i + 1}\nRMSE: {np.sqrt(np.mean(torch_diff**2)):.2e}",
            pad=10,
            fontsize=9,
        )

        # Remove tick labels for cleaner look
        axes[0, i].set_xticks([])
        axes[0, i].set_yticks([])
        axes[1, i].set_xticks([])
        axes[1, i].set_yticks([])

    plt.tight_layout()
    return fig


def plot_error_distributions(original_dms, scipy_coords, torch_coords):
    """
    Create a histogram comparing error distributions for both methods.
    """
    scipy_errors = []
    torch_errors = []

    for i in range(len(original_dms)):
        scipy_dm = squareform(pdist(scipy_coords[i]))
        torch_dm = squareform(pdist(torch_coords[i]))

        scipy_diff = (original_dms[i] - scipy_dm).flatten()
        torch_diff = (original_dms[i] - torch_dm).flatten()

        scipy_errors.extend(scipy_diff)
        torch_errors.extend(torch_diff)

    fig, ax = plt.subplots(figsize=(8, 5))

    # Modern blue and green
    colors = ["#3498db", "#2ecc71"]

    # Plot histograms with improved aesthetics
    ax.hist(
        scipy_errors,
        bins=50,
        alpha=0.6,
        label="Scipy MDS",
        density=True,
        color=colors[0],
        edgecolor="white",
        linewidth=0.5,
    )
    ax.hist(
        torch_errors,
        bins=50,
        alpha=0.6,
        label="Torch SMACOF",
        density=True,
        color=colors[1],
        edgecolor="white",
        linewidth=0.5,
    )

    ax.set_xlabel("Mean Error (Å)", fontsize=10)
    ax.set_ylabel("Density", fontsize=10)
    ax.set_title("Distribution of Errors", fontsize=12, pad=15)

    # Clean up the plot
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.grid(False)

    # Add statistics to legend with cleaner formatting
    ax.legend(
        [
            f"Scipy MDS (μ={np.mean(scipy_errors):.2e})",
            f"Torch SMACOF (μ={np.mean(torch_errors):.2e})",
        ],
        frameon=True,
        framealpha=0.9,
        edgecolor="none",
    )

    return fig


def benchmark_methods(
    dms, n_iter=100, tol=1e-4, n_repeats=3, batch_size=None, verbose=False
):
    """Benchmark methods with batch processing support"""
    scipy_times = []
    torch_times = []

    for _ in range(n_repeats):
        # Time torch SMACOF with batch processing
        start_time = time.perf_counter()
        torch_coords, _ = distance_matrix_to_3d_structure_torch_mds(
            dms,
            # batch=100,
            n_iter=n_iter,
            tol=tol,
        )
        if verbose:
            print("torch", time.perf_counter() - start_time)
        torch_times.append(time.perf_counter() - start_time)

        # Time Scipy MDS
        start_time = time.perf_counter()
        scipy_mds_coords = [distance_matrix_to_3d_structure_mds(dm) for dm in dms]
        scipy_times.append(time.perf_counter() - start_time)

        if verbose:
            print("scipy", scipy_times[-1])

    return {
        "scipy_times": np.array(scipy_times),
        "torch_times": np.array(torch_times),
        "scipy_coords": scipy_mds_coords,
        "torch_coords": torch_coords,
    }


def plot_timing_comparison(timing_results, n_conformers):
    """
    Create a violin plot comparing the timing distributions.

    Args:
        timing_results: Dictionary containing timing measurements
        n_conformers: Number of conformers processed

    Returns:
        matplotlib figure
    """

    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 5))

    # Modern blue and green
    colors = ["#3498db", "#2ecc71"]

    # Create violin plot
    parts = ax1.violinplot(
        [timing_results["scipy_times"], timing_results["torch_times"]], showmeans=True
    )

    # Set colors for individual violins
    for i, body in enumerate(parts["bodies"]):
        body.set_facecolor(colors[i])
        body.set_edgecolor(colors[i])
        body.set_alpha(0.7)  # Optional transparency for better visibility

    # Set the mean line color to red
    parts["cmeans"].set_color("#e74c3c")

    ax1.set_xticks([1, 2])
    ax1.set_xticklabels(["Scipy MDS", "Torch SMACOF"])
    ax1.set_ylabel("Time (seconds)")
    ax1.set_title("Computation Time Distribution", pad=15)

    # Remove top and right spines
    ax1.spines["top"].set_visible(False)
    ax1.spines["right"].set_visible(False)

    # Add mean times as text
    mean_scipy = np.mean(timing_results["scipy_times"])
    mean_torch = np.mean(timing_results["torch_times"])

    speedup_torch = mean_scipy / mean_torch
    text = (
        f"Mean Times:\nScipy: {mean_scipy:.3f}s\ntorch: {mean_torch:.3f}s\n\n"
        f"Speedup (torch): {speedup_torch:.2f}x\n"
        f"Conformers: {n_conformers}"
    )

    ax1.text(
        0.95,
        0.95,
        text,
        transform=ax1.transAxes,
        verticalalignment="top",
        horizontalalignment="right",
        bbox=dict(boxstyle="round", facecolor="white", alpha=0.8),
    )

    # Bar plot with improved aesthetics
    times_per_conformer = [
        np.mean(timing_results["scipy_times"]) / n_conformers,
        np.mean(timing_results["torch_times"]) / n_conformers,
    ]

    bars = ax2.bar(
        ["Scipy MDS", "Torch SMACOF"],
        times_per_conformer,
        color=colors,
        alpha=0.8,
        width=0.6,
    )

    ax2.set_ylabel("Time per Conformer (seconds)")
    ax2.set_title("Average Computation Time", pad=15)

    # Remove top and right spines
    ax2.spines["top"].set_visible(False)
    ax2.spines["right"].set_visible(False)
    ax2.grid(False)

    # Add value labels on bars
    for bar in bars:
        height = bar.get_height()
        ax2.text(
            bar.get_x() + bar.get_width() / 2.0,
            height,
            f"{height:.3f}s",
            ha="center",
            va="bottom",
            fontsize=9,
        )

    plt.tight_layout()
    return fig


def run_benchmark_comparison(dms, n_iter=100, tol=1e-4, n_repeats=5):
    """Run complete benchmark with all visualizations"""
    timing_results = benchmark_methods(dms, n_iter=n_iter, tol=tol, n_repeats=n_repeats)

    # Create all visualization figures
    heatmap_fig = visualize_comparison(
        dms[:3],
        timing_results["scipy_coords"][:3],
        timing_results["torch_coords"][:3],
    )

    dist_fig = plot_error_distributions(
        dms,
        timing_results["scipy_coords"],
        timing_results["torch_coords"],
    )

    timing_fig = plot_timing_comparison(timing_results, len(dms))

    # Save figures
    heatmap_fig.savefig(
        "mds_comparison_sample_heatmaps.pdf", dpi=300, bbox_inches="tight"
    )
    heatmap_fig.savefig(
        "mds_comparison_sample_heatmaps.png", dpi=300, bbox_inches="tight"
    )
    dist_fig.savefig("mds_mean_error_distribution.pdf", dpi=300, bbox_inches="tight")
    dist_fig.savefig("mds_mean_error_distribution.png", dpi=300, bbox_inches="tight")
    timing_fig.savefig("mds_timing_comparison.pdf", dpi=300, bbox_inches="tight")
    timing_fig.savefig("mds_timing_comparison.png", dpi=300, bbox_inches="tight")

    return heatmap_fig, dist_fig, timing_fig, timing_results


if __name__ == "__main__":
    # shape (200, 384, 384)
    ens = generate("PKGS" + "PLKE" * 95)
    dms = ens["sequence_1"].distance_maps()

    heatmap_fig, dist_fig, timing_fig, timing_results = run_benchmark_comparison(
        dms, n_iter=300, tol=1e-4, n_repeats=10
    )



import hdf5plugin
import pytorch_lightning as pl
import torch

from starling.data.data_wrangler import load_hdf5_compressed, read_tsv_file


class MatrixDataset(torch.utils.data.Dataset):
    def __init__(self, tsv_file: str) -> None:
        """
        A class that creates a dataset of distance maps compatible with PyTorch
        tsv_file : str
            A path to a tsv file containing the paths to distance maps as a first column
            and index of a distance map to load as a second column
        """
        self.data = read_tsv_file(tsv_file)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        data_path, frame = self.data.iloc[index]

        distance_map = load_hdf5_compressed(
            data_path, keys_to_load=["dm"], frame=int(frame)
        )["dm"]

        # Add a channel dimension using unsqueeze
        distance_map = torch.from_numpy(distance_map).unsqueeze(0)

        return distance_map


# Step 2: Create a data module
class MatrixDataModule(pl.LightningDataModule):
    def __init__(
        self,
        train_data=None,
        val_data=None,
        test_data=None,
        batch_size=None,
        num_workers=None,
    ):
        super().__init__()
        self.train_data = train_data
        self.val_data = val_data
        self.test_data = test_data
        self.batch_size = batch_size
        self.num_workers = num_workers  # Set this to 16

    def prepare_data(self):
        # Implement any data download or preprocessing here
        pass

    def setup(self, stage: str):
        if stage == "fit":
            self.train_dataset = MatrixDataset(
                self.train_data,
            )
            self.val_dataset = MatrixDataset(
                self.val_data,
            )
        if stage == "test":
            self.test_dataset = MatrixDataset(
                self.test_data,
            )
        if stage == "predict":
            self.predict_dataset = MatrixDataset(
                self.predict_data,
            )

    def train_dataloader(self):
        return torch.utils.data.DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers,
            pin_memory=True,
            persistent_workers=True,
            prefetch_factor=5,
        )

    def val_dataloader(self):
        return torch.utils.data.DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=True,
            persistent_workers=True,
            prefetch_factor=5,
        )

    def test_dataloader(self):
        return torch.utils.data.DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            persistent_workers=True,
            prefetch_factor=5,
            pin_memory=True,
        )



import argparse
import glob
import os

import pytorch_lightning as pl
import wandb
import yaml
from pytorch_lightning.callbacks import LearningRateMonitor, ModelCheckpoint
from pytorch_lightning.loggers import WandbLogger

from starling.data.argument_parser import get_vae_params
from starling.data.VAE_loader import MatrixDataModule
from starling.models.vae import VAE


def train_vae():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--config_file",
        type=str,
        default=None,
        help="Path to the configuration file to use",
    )

    parser.add_argument(
        "--num_nodes",
        type=int,
        default=1,
        help="Number of nodes to use for training",
    )

    parser.add_argument(
        "--num_workers",
        type=int,
        default=16,
        help="Number of workers to use for loading in the data",
    )

    args = parser.parse_args()

    # Reads in default and user defined configuration arguments
    config = get_vae_params(config_file=args.config_file)

    # Make the directories to save the model and logs
    os.makedirs(config["training"]["output_path"], exist_ok=True)

    with open(f"{config['training']['output_path']}/config.yaml", "w") as f:
        yaml.dump(config, f)

    # Set up model checkpoint saving
    checkpoint_callback = ModelCheckpoint(
        monitor="epoch_val_loss",  # Monitor validation loss for saving the best model
        dirpath=f"{config['training']['output_path']}/",  # Directory to save checkpoints
        filename="model-kernel-{epoch:02d}-{epoch_val_loss:.2f}",  # File name format for saved models
        save_top_k=-1,
        mode="min",  # Minimize the monitored metric (val_loss)
    )

    # delta = datetime.timedelta(hours=1)
    save_last_checkpoint = ModelCheckpoint(
        dirpath=f"{config['training']['output_path']}/",  # Directory to save checkpoints
        filename="last",
    )

    checkpoint_dir = f"{config['training']['output_path']}/"
    checkpoint_pattern = os.path.join(checkpoint_dir, "last.ckpt")

    # Check if any checkpoint exists
    checkpoint_files = glob.glob(checkpoint_pattern)

    if checkpoint_files:
        # This will load the most recent checkpoint
        ckpt_path = "last"
    else:
        # This will start training from scratch
        ckpt_path = None

    lr_monitor = LearningRateMonitor(logging_interval="step")

    # Set up data loaders
    dataset = MatrixDataModule(
        **config["data"],
        num_workers=args.num_workers,
    )

    dataset.setup(stage="fit")

    vae = VAE(**config["model"])

    with open(f"{config['training']['output_path']}/model_architecture.txt", "w") as f:
        f.write(str(vae))
    ##############################

    # Set up logging on weights and biases
    wandb_logger = WandbLogger(project=config["training"]["project_name"])
    wandb_logger.watch(vae)

    # Set up PyTorch Lightning Trainer
    trainer = pl.Trainer(
        accelerator="auto",
        devices=config["device"]["cuda"],
        num_nodes=args.num_nodes,
        max_epochs=config["training"]["num_epochs"],
        callbacks=[checkpoint_callback, lr_monitor, save_last_checkpoint],
        # gradient_clip_val=1.0,
        precision="bf16-mixed",
        logger=wandb_logger,
    )

    # Start training
    trainer.fit(vae, dataset, ckpt_path=ckpt_path)

    # Detach the logging on wandb
    wandb_logger.experiment.unwatch(vae)
    wandb.finish()


if __name__ == "__main__":
    train_vae()



import gc
import os
import time
from datetime import datetime

import numpy as np
import torch
from soursop.sstrajectory import SSTrajectory
from tqdm import tqdm

from starling import configs
from starling.inference.model_loading import ModelManager
from starling.samplers.ddim_sampler import DDIMSampler
from starling.structure.coordinates import (
    create_ca_topology_from_coords,
    generate_3d_coordinates_from_distances,
)
from starling.structure.ensemble import Ensemble

# initialize model_manager singleton. This happens when this module
# is imported to ensemble_generation, so we can use the
# same model_manager for all calls to generate_backend.
model_manager = ModelManager()


def symmetrize_distance_map(dist_map):
    """
    Symmetrize a distance map by replacing the lower triangle with the upper triangle values.

    Parameters
    ----------
    dist_map : torch.Tensor
        A 2D tensor representing the distance map.

    Returns
    -------
    torch.Tensor
        A symmetrized distance map.

    """

    # Ensure the distance map is 2D
    dist_map = dist_map.squeeze(0) if dist_map.dim() == 3 else dist_map

    # Create a copy of the distance map to modify
    sym_dist_map = dist_map.clone()

    # Replace the lower triangle with the upper triangle values
    mask_upper_triangle = torch.triu(torch.ones_like(dist_map), diagonal=1).bool()
    mask_lower_triangle = ~mask_upper_triangle

    # Set lower triangle values to be the same as the upper triangle
    sym_dist_map[mask_lower_triangle] = dist_map.T[mask_lower_triangle]

    # Set diagonal values to zero
    sym_dist_map.fill_diagonal_(0)

    return sym_dist_map.cpu()


def generate_backend(
    sequence_dict,
    conformations,
    device,
    steps,
    ddim,
    return_structures,
    batch_size,
    num_cpus_mds,
    num_mds_init,
    output_directory,
    return_data,
    verbose,
    show_progress_bar,
    show_per_step_progress_bar,
    pdb_trajectory,
    model_manager=model_manager,
):
    """
    Backend function for generating the distance maps using STARLING.

    NOTE - this function does VERY littel sanity checking; to actually perform
    predictions use starling.frontend.ensemble_generation. This is NOT
    a user facing function!

    Parameters
    ---------------
    sequence_dict : dict
        A dictionary with the sequence names as the key and the
       sequences as the values. These names will be used to write
       any output files (if writing is requested).

    ddpm : str
        The path to the DDPM model

    device : str
        The device to use for predictions.

    steps : int
        The number of steps to run the DDPM model.

    ddim : bool
        Whether to use DDIM for sampling.

    return_structures : bool
        Whether to return the 3D structure.

    batch_size : int
        The batch size to use for sampling.

    num_cpus_mds : int
        The number of CPUs to use for MDS. There
        is no point specifying more than the default
        number of MDS runs performed (defined in configs)

    output_directory : str or None
        If None, no output is saved.
        If not None, will save the output to the specified path.
        This includes the distance maps and if return_structures=True,
        the 3D structures.
        The distance maps are saved as .npy files with the names
        <sequence_name>_STARLING_DM.npy
        and the structures are save with the file names
        <sequence_name>_STARLING.xtc and <sequence_name>_STARLING.pdb.

    return_data : bool
        If True, will return the distance maps and structures (if generated)
        as a dictionary regardless of the output_directory. If False, will
        return None. Note the reason to set this to None is if you're
        predicting a large set of sequences this will save memory.

    verbose : bool
        Whether to print verbose output. Default is False.

    show_progress_bar : bool
        Whether to show a progress bar. Default is True.

    show_per_step_progress_bar : bool, optional
        whether to show progress bar per step.
    pdb_trajectory: bool
        Whether to save the trajectory as a PDB file. Default is False.

    model_manager : ModelManager
        A ModelManager object to manage loaded models.
        This lets us avoid loading the model iteratively
        when calling generate multiple times in a single
        session. Default is model_manager, which is initialized
        outside of this function code block. To update the path
        to the models, update the paths in config.py, which are
        read into the ModelManager object located the
        model_loading.py

    Returns
    ---------------
    dict or None:
        A dict with the sequence names as the key and
        a starling.ensembl.Ensemble objects for each
        sequence as values.

        If output_directory is not none, the output will save to
        the specified path.
    """

    overall_start_time = time.time()

    # get models. This will only load once even if we call this
    # function multiple times.
    encoder_model, diffusion = model_manager.get_models(device=device)

    # Construct a sampler
    if ddim:
        sampler = DDIMSampler(ddpm_model=diffusion, n_steps=steps)
    else:
        sampler = diffusion

    # get num_batchs and remaining samples
    num_batches = conformations // batch_size
    remaining_samples = conformations % batch_size

    # dictionary to hold distance maps and structures if applicable.
    output_dict = {}

    # see if a progress bar is wanted. If it is, set it up.
    if show_progress_bar:
        pbar = tqdm(total=len(sequence_dict))

    # iterate over sequence_dict
    for num, seq_name in enumerate(sequence_dict):
        ## -----------------------------------------
        ## Start of prediction cycle for this sequence

        start_time_prediction = time.time()

        # list to hold distance maps
        starling_dm = []

        # get sequence
        sequence = sequence_dict[seq_name]

        # iterate over batches for actual DDIM sampling
        for batch in range(num_batches):
            distance_maps = sampler.sample(
                batch_size,
                labels=sequence,
                show_per_step_progress_bar=show_per_step_progress_bar,
            )
            starling_dm.append(
                [
                    symmetrize_distance_map(dm[:, : len(sequence), : len(sequence)])
                    for dm in distance_maps
                ]
            )

        # iterate over remaining samples
        if remaining_samples > 0:
            distance_maps = sampler.sample(
                remaining_samples,
                labels=sequence,
                show_per_step_progress_bar=show_per_step_progress_bar,
            )
            starling_dm.append(
                [
                    symmetrize_distance_map(dm[:, : len(sequence), : len(sequence)])
                    for dm in distance_maps
                ]
            )

        # concatenate symmetrized distance maps.
        sym_distance_maps = torch.cat(
            [torch.stack(batch) for batch in starling_dm], dim=0
        )

        end_time_prediction = time.time()

        ##
        ## NB: We keep the ensemble reconstruction code here for now because
        ## at this point the distance maps are still tensors, not np.arrays
        ## so we avoid unnecessary conversions by using the tensor versions
        ## here.
        ##
        # if return_structures is True, generate 3D structure

        # set time at which we start structure generation to 0
        start_time_structure_generation = time.time()

        # we initialize this to 0 and will update as needed (or not)
        end_time_structure_generation = time.time()

        if return_structures:
            coordinates = generate_3d_coordinates_from_distances(
                device,
                batch_size,
                num_cpus_mds,
                num_mds_init,
                sym_distance_maps,
                progress_bar=show_progress_bar,
            )

            # make traj as an sstrajectory object and extract out the ssprotein object
            ssprotein = SSTrajectory(
                TRJ=create_ca_topology_from_coords(sequence, coordinates)
            ).proteinTrajectoryList[0]

            end_time_structure_generation = time.time()

        # if no structures are requested, set ssprotein to None
        else:
            ssprotein = None

        # pull the distance maps out of the tensor and convert to numpy
        final_distance_maps = sym_distance_maps.detach().cpu().numpy()

        # create Ensemble object. Note if the ssprotein argument is None
        # this is expected and will initialize the ensemble without
        # structures
        E = Ensemble(final_distance_maps, sequence, ssprot_ensemble=ssprotein)

        # if we are saving things, save as we progress through so we generate
        # structures/DMs in situ
        if output_directory is not None:
            # num == 0 just means we are on the first sequence.
            if verbose and num == 0:
                print(f"Saving results to: {os.path.abspath(output_directory)}")

            # if we're saving structures do that first;
            if return_structures:
                # this saves both a topology (PDB) and a trajectory (XTC) file
                E.save_trajectory(
                    filename_prefix=os.path.join(
                        output_directory, seq_name + "_STARLING"
                    ),
                    pdb_trajectory=pdb_trajectory,
                )

            # save full ensemble
            E.save(os.path.join(output_directory, f"{seq_name}"))

        ## End of prediction cycle for this sequence
        ## -----------------------------------------

        # if we are returning data, add the data to the output_dict
        if return_data:
            output_dict[seq_name] = E

        # if not, force cleanup of things to save memory
        else:
            del E
            del final_distance_maps
            gc.collect()

        # update progress bar if we have one.
        if show_progress_bar:
            pbar.update(1)

        if verbose:
            elapsed_time_structure_generation = (
                end_time_structure_generation - start_time_structure_generation
            )
            elapsed_time_prediction = end_time_prediction - start_time_prediction
            total_time = elapsed_time_structure_generation + elapsed_time_prediction
            n_conformers = len(sym_distance_maps)

            print(
                f"\n\n##### SUMMARY OF SEQUENCE PREDICTION ({num + 1}/{len(sequence_dict)}) #####"
            )
            print(f"Sequence name                       : {seq_name}")
            print(f"Sequence length                     : {len(sequence)}")
            print(f"Number of conformers                : {n_conformers}")
            print(f"Number of steps                     : {steps}")
            print(
                f"Total time for prediction           : {round(elapsed_time_prediction, 2)}s ({round(100 * (elapsed_time_prediction / total_time), 2)}% of time)"
            )
            print(
                f"Total time for structure generation : {round(elapsed_time_structure_generation, 2)}s ({round(100 * (elapsed_time_structure_generation / total_time), 2)}% of time)"
            )
            print(f"Time per conformer                  : {total_time / n_conformers}s")
            print("\n")
        else:
            # changed from print to pass because if not verbose, we don't need to do anything.
            pass

    # make sure we close the progress bar if we used one
    if show_progress_bar:
        pbar.close()

    if verbose:
        # Convert total time to hours, minutes, and seconds
        overall_time = time.time() - overall_start_time
        total_hours = round(overall_time // 3600, 2)
        total_minutes = round((overall_time % 3600) // 60, 2)
        total_seconds = round(overall_time % 60, 2)
        print("-------------------------------------------------------")
        print(f"Summary of all predictions ({len(sequence_dict)} sequences)")
        print("-------------------------------------------------------")
        print(
            f"\nTotal time (all sequences, all I/O) : {total_hours} hrs {total_minutes} mins {total_seconds} secs"
        )

        current_datetime = datetime.now()
        formatted_datetime = current_datetime.strftime("%Y-%m-%d %H:%M:%S")

        print("STARLING predictions completed at:", formatted_datetime)
        print("")

    return output_dict



import torch
from torch import nn


# Non-learnable position encodings
class PositionalEncoding1D(nn.Module):
    def __init__(self, max_seq_len, embedding_size):
        """
        Positional encoding for 1D data. The positional encoding is added to the input tensor
        to provide information about the position of the elements in the input data. The positional
        encoding is computed using sine and cosine functions.

        Parameters
        ----------
        max_seq_len : _type_
            Max sequence length of the input data.
        embedding_size : _type_
            The number of features of the input data.
        """
        super(PositionalEncoding1D, self).__init__()
        self.max_seq_len = max_seq_len
        self.embedding_size = embedding_size
        self.positional_encoding = self._generate_positional_encoding()

    def _generate_positional_encoding(self):
        # Initialize the positional encoding tensor with 0s
        pe = torch.zeros(self.max_seq_len, self.embedding_size)

        # Get the position tensor (0, 1, 2, ..., max_seq_len - 1)
        position = torch.arange(0, self.max_seq_len, dtype=torch.float32).unsqueeze(1)

        # Compute divisor term for the positional encodings
        div_term = torch.exp(
            torch.arange(0, self.embedding_size, 2, dtype=torch.float32)
            * (-torch.log(torch.tensor(10000.0)) / self.embedding_size)
        )

        # Assigns sine values to even indices in the last dimension
        pe[:, 0::2] = torch.sin(position * div_term)

        # Assigns cosine values to odd indices in the last dimension
        pe[:, 1::2] = torch.cos(position * div_term)

        # Add batch dimension
        pe = pe.unsqueeze(0)

        return pe

    def forward(self, x):
        if self.positional_encoding.device != x.device:
            self.positional_encoding = self.positional_encoding.to(x.device)

        # Add positional encoding to the input tensor
        return x + self.positional_encoding


class PositionalEncoding2D(nn.Module):
    def __init__(self, embed_dim: int):
        """
        Positional encoding for 2D data. The positional encoding is added to the input tensor
        to provide information about the position of the elements in the input data. The positional
        encoding is computed using sine and cosine functions.

        Parameters
        ----------
        embed_dim : int
            The number of features of the input data.
        """
        super(PositionalEncoding2D, self).__init__()

        self.embed_dim = embed_dim

    def forward(self, x):
        b, c, h, w = x.shape
        pe = self.generate_pe(h, w, x.device)
        return x + pe

    def generate_pe(self, height, width, device):
        # Initialize the positional encoding tensor with 0s
        pe = torch.zeros(self.embed_dim, height, width, device=device)

        # Get the position tensors for height and width of the 2D data
        y_position = torch.arange(
            0, height, dtype=torch.float32, device=device
        ).unsqueeze(1)
        x_position = torch.arange(
            0, width, dtype=torch.float32, device=device
        ).unsqueeze(0)

        # Compute divisor term for the positional encodings
        div_term = torch.exp(
            torch.arange(0, self.embed_dim, 2, dtype=torch.float32, device=device)
            * (-torch.log(torch.tensor(10000.0, device=device)) / self.embed_dim)
        )

        # Reshape div_term to match (embed_dim/2, height, width) for broadcasting
        div_term = div_term.unsqueeze(1).unsqueeze(1)  # Shape (embed_dim/2, 1, 1)

        # Compute the positional encodings for height and width
        pe_x = torch.sin(x_position.unsqueeze(0) * div_term)
        pe_y = torch.sin(y_position.unsqueeze(0) * div_term)

        # Assign to the positional encoding tensor (even indices for x, odd indices for y)
        pe[0::2, :, :] = pe_x
        pe[1::2, :, :] = pe_y

        # Return and add batch dimension
        return pe.unsqueeze(0)


# Learnable positional encodings
class LearnablePositionalEncoding1D(nn.Module):
    def __init__(self, sequence_length, embed_dim):
        super(LearnablePositionalEncoding1D, self).__init__()
        self.sequence_length = sequence_length
        self.embed_dim = embed_dim
        self.positional_encoding = nn.Parameter(
            torch.randn(1, sequence_length, embed_dim)
        )

    def forward(self, x):
        if self.positional_encoding.device != x.device:
            self.positional_encoding = self.positional_encoding.to(x.device)

        return x + self.positional_encoding


class LearnablePositionalEncoding2D(nn.Module):
    def __init__(self, height, width, embed_dim):
        super(LearnablePositionalEncoding2D, self).__init__()
        self.height = height
        self.width = width
        self.embed_dim = embed_dim
        self.positional_encoding = nn.Parameter(
            torch.randn(1, embed_dim, height, width)
        )

    def forward(self, x):
        if self.positional_encoding.device != x.device:
            self.positional_encoding = self.positional_encoding.to(x.device)

        return x + self.positional_encoding



import hdf5plugin
import numpy as np
import pytorch_lightning as pl
import torch

from starling.data.data_wrangler import (
    load_hdf5_compressed,
    read_tsv_file,
)


def sequence_to_indices(sequence, aa_to_int):
    """
    Converts a single sequence to integer indices based on a mapping.

    Parameters:
    - sequence (str): A single sequence string.
    - aa_to_int (dict): A dictionary mapping amino acids to integers.

    Returns:
    - torch.Tensor: Tensor of integer indices.
    """
    int_sequence = [aa_to_int[aa] for aa in sequence]
    return torch.tensor(int_sequence, dtype=torch.int64)


class MatrixDataset(torch.utils.data.Dataset):
    def __init__(self, tsv_file: str, labels: str) -> None:
        """
        A class that creates a dataset of distance maps compatible with PyTorch
        tsv_file : str
            A path to a tsv file containing the paths to distance maps as a first column
            and index of a distance map to load as a second column
        labels : str
            Which labels to use for the dataset, learnable or fixed (finches interaction matrix).
        """
        self.data = read_tsv_file(tsv_file)
        self.labels = labels
        self.aa_to_int = {
            "0": 0,
            "A": 1,
            "C": 2,
            "D": 3,
            "E": 4,
            "F": 5,
            "G": 6,
            "H": 7,
            "I": 8,
            "K": 9,
            "L": 10,
            "M": 11,
            "N": 12,
            "P": 13,
            "Q": 14,
            "R": 15,
            "S": 16,
            "T": 17,
            "V": 18,
            "W": 19,
            "Y": 20,
        }

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        data_path, frame = self.data.iloc[index]
        data = load_hdf5_compressed(
            data_path, keys_to_load=["dm", "seq"], frame=int(frame)
        )
        distance_map = data["dm"].astype(np.float16)
        distance_map = torch.from_numpy(distance_map).unsqueeze(0)

        sequence = data["seq"].astype(np.int32)
        sequence = torch.from_numpy(sequence)

        return distance_map, sequence


# Step 2: Create a data module
class MatrixDataModule(pl.LightningDataModule):
    def __init__(
        self,
        train_data=None,
        val_data=None,
        test_data=None,
        batch_size=None,
        labels=None,
        num_workers=None,
    ):
        super().__init__()
        self.train_data = train_data
        self.val_data = val_data
        self.test_data = test_data
        self.batch_size = batch_size
        self.labels = labels
        # self.num_workers = int(os.cpu_count() / 4)
        self.num_workers = num_workers

    def prepare_data(self):
        # Implement any data download or preprocessing here
        pass

    def setup(self, stage: str):
        if stage == "fit":
            self.train_dataset = MatrixDataset(
                self.train_data,
                labels=self.labels,
            )
            self.val_dataset = MatrixDataset(
                self.val_data,
                labels=self.labels,
            )
        if stage == "test":
            self.test_dataset = MatrixDataset(
                self.test_data,
                labels=self.labels,
            )
        if stage == "predict":
            self.predict_dataset = MatrixDataset(
                self.predict_data,
                labels=self.labels,
            )

    def train_dataloader(self):
        return torch.utils.data.DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers,
            pin_memory=True,
            persistent_workers=True,
            prefetch_factor=5,
        )

    def val_dataloader(self):
        return torch.utils.data.DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            persistent_workers=True,
            prefetch_factor=5,
            pin_memory=True,
        )

    def test_dataloader(self):
        return torch.utils.data.DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=True,
        )



"""Construction of intrinsically disordered proteins ensembles through multiscale generative models"""

# Add imports here
from ._version import __version__

# Import submodules to make them accessible as part of the top-level package
from starling.data import *
from starling.models import *
from starling.training import *

import starling.configs
from starling.frontend.ensemble_generation import generate

from starling.structure.ensemble import load_ensemble







